# Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

# * The software is provided "as is" without any warranty, neither express
#   nor implied.
# * In no event will the author be held liable for any damage(s) arising
#   from the use of the software.
# * Redistribution of the software is allowed only in unmodified form.
# * Permission is granted to use the software solely for personal private
#   and non-commercial purposes.
# * An individuals use of the software in his or her capacity or function
#   as an agent, (independent) contractor, employee, member or officer of
#   a business, corporation or organization (commercial or non-commercial)
#   does not qualify as personal private and non-commercial purpose.
# * Without written approval from the author the software must not be used
#   for a business, for commercial, corporate, governmental, military or
#   organizational purposes of any kind, or in a commercial, corporate,
#   governmental, military or organizational environment of any kind.

.SUFFIXES:

all: AMD64-MSC.LIB

clean:
	ERASE AMD64-MSC.DOS AMD64-MSC.EXP AMD64-MSC.OBJ
	ERASE AMD64-CRT.OBJ AMD64-DLG.OBJ AMD64-DLH.OBJ AMD64-DLI.OBJ AMD64-FMT.OBJ AMD64-GS.OBJ AMD64-MEM.OBJ AMD64-RTC.OBJ AMD64-STK.OBJ AMD64-STR.OBJ AMD64-TLS.OBJ AMD64-WCS.OBJ

clobber: clean
	ERASE AMD64-MSC.EXE AMD64-MSC.LIB

test: AMD64-MSC.DOS AMD64-MSC.LIB AMD64-MSC.EXE
	"$(MAKEDIR)\AMD64-MSC.EXE"

AMD64-MSC.IMP: AMD64-MSC.MAK
	LINK.EXE /LIB /BREPRO /DEF /EXPORT:__C_specific_handler /EXPORT:__chkstk /EXPORT:memchr /EXPORT:memcmp /EXPORT:memcpy /EXPORT:memmove /EXPORT:memset /MACHINE:AMD64 /NAME:NTDLL /NODEFAULTLIB /NOLOGO /OUT:$@

AMD64-MSC.LIB: AMD64-CRT.OBJ AMD64-DLI.OBJAMD64-FMT.OBJ AMD64-GS.OBJ AMD64-MEM.OBJ AMD64-STK.OBJ AMD64-STR.OBJ AMD64-TLS.OBJ AMD64-WCS.OBJ # AMD64-RTC.OBJ
	LINK.EXE /LIB /BREPRO /DEF /EXPORT:__C_specific_handler /MACHINE:AMD64 /NAME:KERNEL32 /NODEFAULTLIB /NOLOGO /OUT:$@
	LINK.EXE /LIB /BREPRO /MACHINE:AMD64 /NODEFAULTLIB /NOLOGO /OUT:$@ $@ $**

AMD64-MSC.EXE: AMD64-MSC.DOS AMD64-MSC.LIB AMD64-MSC.OBJ
	LINK.EXE /LINK /BREPRO /DEFAULTLIB:AMD64-MSC.LIB /DEFAULTLIB:KERNEL32.LIB /DEFAULTLIB:USER32.LIB /ENTRY:wmainCRTStartup /FIXED /MACHINE:AMD64 /MERGE:.rdata=.const /MERGE:.text=.code /NOCOFFGRPINFO /NOLOGO /NXCOMPAT /OPT:ICF,REF /OSVERSION:5.2 /OUT:$@ /RELEASE /SECTION:.bss,d /STUB:AMD64-MSC.DOS /SUBSYSTEM:CONSOLE /SWAPRUN:CD,NET /TEST /VERSION:0.815 AMD64-MSC.OBJ
	LINK.EXE /EDIT /NOLOGO /SECTION:.bss,d $@

AMD64-MSC.DOS: AMD64-MSC.MAK
	CERTUTIL.EXE /DecodeHex /F /V << $@
4D 5A 90 00 01 00 00 00 04 00 00 00 FF FF 00 00   MZ..............
E0 00 00 00 43 00 00 00 40 00 00 00 00 00 00 00   ....C...@.......
00 00 00 00 19 57 04 27 00 00 00 00 00 00 00 00   .....W.'........
00 00 00 00 00 00 00 00 00 00 00 00 00 00 00 00   ................
28 43 29 6F 70 79 72 69 67 68 74 20 32 30 30 34   (C)opyright 2004
2D 32 30 32 34 2C 20 53 74 65 66 61 6E 20 4B 61   -2024, Stefan Ka
6E 74 68 61 6B 20 3C 73 74 65 66 61 6E 2E 6B 61   nthak <stefan.ka
6E 74 68 61 6B 40 6E 65 78 67 6F 2E 64 65 3E 0D   nthak@nexgo.de>.
0A 07 24 0E 1F 33 D2 B4 09 CD 21 B8 01 4C CD 21   .....3....!..L.!
<<NOKEEP

AMD64-MSC.OBJ: AMD64-MSC.MAK
	CL.EXE /Brepro /c /Fo$@ /GAFwy /nologo /O2txy /Tc<< /W4 /Zl
// Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

// https://msdn.microsoft.com/en-us/library/7f0aews7.aspx

#pragma comment(compiler)
#pragma comment(lib, "kernel32.lib")
#pragma comment(user, "(C)opyright 2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>")

#define STRICT
#define UNICODE
#define WIN32_LEAN_AND_MEAN

#include <windows.h>

__declspec(thread)
DWORD	dwTLS = 'MSVC';		// placed in writable ".tls$$" section by the compiler

extern	const	DWORD	_tls_index;

// https://msdn.microsoft.com/en-us/library/hh916382.aspx

__declspec(safebuffers)
BOOL	CDECL	PrintConsole(HANDLE hConsole, [SA_FormatString(Style="printf")] LPCWSTR lpFormat, ...)
{
	WCHAR	szOutput[1024];
	DWORD	dwOutput;
	DWORD	dwConsole;

	// https://msdn.microsoft.com/en-us/library/kb57fad8.aspx

	va_list	vaInput;
	va_start(vaInput, lpFormat);

	// https://msdn.microsoft.com/en-us/library/ms647551.aspx

	dwOutput = wvsprintf(szOutput, lpFormat, vaInput);

	va_end(vaInput);

	if (dwOutput == 0UL)
		return FALSE;

	// https://msdn.microsoft.com/en-us/library/ms687401.aspx

	if (!WriteConsole(hConsole, szOutput, dwOutput, &dwConsole, NULL))
		return FALSE;

	return dwConsole == dwOutput;
}

const	LPCWSTR	szReason[4] = {L"process detach",
		               L"process attach",
		               L"thread attach",
		               L"thread detach"};

// https://msdn.microsoft.com/en-us/library/hh916382.aspx

__declspec(safebuffers)
VOID	WINAPI	TLSCallback(LPVOID hModule, DWORD dwReason, LPVOID lpUnused)
{
	// https://msdn.microsoft.com/en-us/library/ms683231.aspx

	HANDLE	hOutput = GetStdHandle(STD_OUTPUT_HANDLE);

	HMODULE	hCaller;
	DWORD	dwCaller;
	WCHAR	szCaller[MAX_PATH];
	WCHAR	szModule[MAX_PATH];

	// https://msdn.microsoft.com/en-us/library/ms683197.aspx

	DWORD	dwModule = GetModuleFileName(hModule, szModule, sizeof(szModule) / sizeof(*szModule));

	if (hOutput == INVALID_HANDLE_VALUE)
		return;

	if (dwModule < sizeof(szModule) / sizeof(*szModule))
		szModule[dwModule] = '\0';

	// https://msdn.microsoft.com/en-us/library/ms683200.aspx

	if (!GetModuleHandleEx(GET_MODULE_HANDLE_EX_FLAG_FROM_ADDRESS | GET_MODULE_HANDLE_EX_FLAG_UNCHANGED_REFCOUNT,
	                       _ReturnAddress(),
	                       &hCaller))
		szCaller[0] = L'\0';
	else
	{	// https://msdn.microsoft.com/en-us/library/ms683197.aspx

		dwCaller = GetModuleFileName(hCaller, szCaller, sizeof(szCaller) / sizeof(*szCaller));

		if (dwCaller < sizeof(szCaller) / sizeof(*szCaller))
			szCaller[dwCaller] = L'\0';
	}

	PrintConsole(hOutput,
	             L"\n"
	             __LPREFIX(__FUNCTION__) L"() function @ 0x%p\n"
	             L"\tCalled module  @ 0x%p = %ls\n"
	             L"\tCalling module @ 0x%p = %ls\n"
	             L"\tReturn address @ 0x%p = 0x%p\n"
	             L"\tArguments:\n"
	             L"\t\tModule = 0x%p\n"
	             L"\t\tReason = %lu (%ls)\n"
	             L"\t\tUnused = 0x%p\n"
	             L"\tThread id = %lu\n",
	             TLSCallback,
	             hModule, szModule,
	             hCaller, szCaller,
	             _AddressOfReturnAddress(), _ReturnAddress(),
	             hModule, dwReason, szReason[dwReason], lpUnused,
	             GetCurrentThreadId());
}

#pragma const_seg(".ptr$$")

//const	PIMAGE_TLS_CALLBACK	_tls_callback = TLSCallback;

#pragma const_seg()

__declspec(allocate(".ptr$$"))
const	PIMAGE_TLS_CALLBACK	_tls_callback = TLSCallback;

extern	IMAGE_TLS_DIRECTORY	_tls_used;

extern	IMAGE_DOS_HEADER	__ImageBase;

// https://msdn.microsoft.com/en-us/library/hh916382.aspx

__declspec(safebuffers)
DWORD	WINAPI	ThreadProc(LPVOID lpParameter)
{
	HMODULE	hCaller;
	DWORD	dwCaller;
	WCHAR	szCaller[MAX_PATH];
	WCHAR	szModule[MAX_PATH];

	// https://msdn.microsoft.com/en-us/library/ms683197.aspx

	DWORD	dwModule = GetModuleFileName((HMODULE) &__ImageBase, szModule, sizeof(szModule) / sizeof(*szModule));

	// https://msdn.microsoft.com/en-us/library/ms683200.aspx

	if (!GetModuleHandleEx(GET_MODULE_HANDLE_EX_FLAG_FROM_ADDRESS | GET_MODULE_HANDLE_EX_FLAG_UNCHANGED_REFCOUNT,
	                       _ReturnAddress(),
	                       &hCaller))
		szCaller[0] = L'\0';
	else
	{	// https://msdn.microsoft.com/en-us/library/ms683197.aspx

		dwCaller = GetModuleFileName(hCaller, szCaller, sizeof(szCaller) / sizeof(*szCaller));

		if (dwCaller < sizeof(szCaller))
			szCaller[dwCaller] = L'\0';
	}

	if (dwModule < sizeof(szModule) / sizeof(*szModule))
		szModule[dwModule] = L'\0';

	PrintConsole(lpParameter,
	             L"\n"
	             __LPREFIX(__FUNCTION__) L"() function @ 0x%p\n"
	             L"\tCalled module  @ 0x%p = %ls\n"
	             L"\tCalling module @ 0x%p = %ls\n"
	             L"\tReturn address @ 0x%p = 0x%p\n"
	             L"\tParameter = 0x%p\n"
	             L"\tThread id = %lu\n",
	             ThreadProc,
	             &__ImageBase, szModule,
	             hCaller, szCaller,
	             _AddressOfReturnAddress(), _ReturnAddress(),
	             lpParameter,
	             GetCurrentThreadId());

	// https://msdn.microsoft.com/en-us/library/ms682659.aspx
	// https://msdn.microsoft.com/en-us/library/ms679360.aspx

	ExitThread(GetLastError());
}

// https://msdn.microsoft.com/en-us/library/f9t8842e.aspx

__declspec(noreturn)
VOID	CDECL	wmainCRTStartup(VOID)
{
	DWORD	dwError = ERROR_SUCCESS;

	// https://msdn.microsoft.com/en-us/library/ms683182.aspx

	DWORD	dwThreadId = GetCurrentThreadId();

	DWORD	dwThread;
	HANDLE	hThread;

	// https://msdn.microsoft.com/en-us/library/ms683231.aspx

	HANDLE	hOutput = GetStdHandle(STD_OUTPUT_HANDLE);

	HMODULE	hCaller;
	DWORD	dwCaller;
	WCHAR	szCaller[MAX_PATH];
	WCHAR	szModule[MAX_PATH];

	// https://msdn.microsoft.com/en-us/library/ms683197.aspx

	DWORD	dwModule = GetModuleFileName((HMODULE) &__ImageBase, szModule, sizeof(szModule) / sizeof(*szModule));

	if (hOutput == INVALID_HANDLE_VALUE)
		dwError = GetLastError();
	else
	{	// https://msdn.microsoft.com/en-us/library/ms683200.aspx

		if (!GetModuleHandleEx(GET_MODULE_HANDLE_EX_FLAG_FROM_ADDRESS | GET_MODULE_HANDLE_EX_FLAG_UNCHANGED_REFCOUNT,
		                       _ReturnAddress(),
		                       &hCaller))
			szCaller[0] = L'\0';
		else
		{	// https://msdn.microsoft.com/en-us/library/ms683197.aspx

			dwCaller = GetModuleFileName(hCaller, szCaller, sizeof(szCaller) / sizeof(*szCaller));

			if (dwCaller < sizeof(szCaller) / sizeof(*szCaller))
				szCaller[dwCaller] = L'\0';
		}

		if (dwModule < sizeof(szModule) / sizeof(*szModule))
			szModule[dwModule] = L'\0';

		PrintConsole(hOutput,
		             L"\a\n"
		             __LPREFIX(__FUNCTION__) L"() function @ 0x%p\n"
		             L"\tCalled module  @ 0x%p = %ls\n"
		             L"\tCalling module @ 0x%p = %ls\n"
		             L"\tReturn address @ 0x%p = 0x%p\n"
		             L"\tThread id = %lu\n"
		             L"\tTLS index = %ld\n"
		             L"\tTLS value = 0x%p\n"
		             L"\tTLS array @ 0x%p\n"
		             L"\tTLS block @ 0x%p\n"
		             L"\tTLS dword @ 0x%p = \"%.4hs\"\n"
		             L"\tTLS directory     @ 0x%p\n"
		             L"\t\tStart     @ 0x%p\n"
		             L"\t\tEnd       @ 0x%p\n"
		             L"\t\tIndex     @ 0x%p\n"
		             L"\t\tCallbacks @ 0x%p\n"
		             L"\t\tZerofill  = 0x%08lX = \"%.4hs\"\n"
		             L"\t\tAlignment = 0x%08lX\n" + (dwTLS == 'MSVC'),
		             wmainCRTStartup,
		             &__ImageBase, szModule,
		             hCaller, szCaller,
		             _AddressOfReturnAddress(), _ReturnAddress(),
		             dwThreadId,
		             _tls_index,
		             TlsGetValue(_tls_index),
#ifdef _M_IX86
		             __readfsdword(44),
		             ((LPVOID *) __readfsdword(44))[_tls_index],
#elif _M_AMD64
		             __readgsqword(88),
		             ((LPVOID *) __readgsqword(88))[_tls_index],
#else
#error Only I386 and AMD64 supported!
#endif
		             &dwTLS, &dwTLS,
		             &_tls_used,
		             _tls_used.StartAddressOfRawData,
		             _tls_used.EndAddressOfRawData,
		             _tls_used.AddressOfIndex,
		             _tls_used.AddressOfCallBacks,
		             _tls_used.SizeOfZeroFill, &_tls_used.SizeOfZeroFill,
		             _tls_used.Characteristics);

		// https://msdn.microsoft.com/en-us/library/ms682453.aspx

		hThread = CreateThread((LPSECURITY_ATTRIBUTES) NULL,
		                       (SIZE_T) 65536,
		                       ThreadProc,
		                       hOutput,
		                       0UL,
		                       &dwThreadId);

		if (hThread == NULL)
			PrintConsole(hOutput,
			             L"CreateThread() returned error %lu\n",
			             dwError = GetLastError());
		else
		{
			PrintConsole(hOutput,
			             L"\n"
			             L"Thread %lu created and started\n",
			             dwThreadId);

			// https://msdn.microsoft.com/en-us/library/ms687032.aspx

			if (WaitForSingleObject(hThread, INFINITE) == WAIT_FAILED)
				PrintConsole(hOutput,
				             L"WaitForSingleObject() returned error %lu\n",
				             dwError = GetLastError());

			// https://msdn.microsoft.com/en-us/library/ms683190.aspx

			if (!GetExitCodeThread(hThread, &dwThread))
				PrintConsole(hOutput,
				             L"GetExitCodeThread() returned error %lu\n",
				             dwError = GetLastError());
			else
				PrintConsole(hOutput,
				             L"\n"
				             L"Thread %lu exited with code %lu\n",
				             dwThreadId, dwThread);

			// https://msdn.microsoft.com/en-us/library/ms724211.aspx

			if (!CloseHandle(hThread))
				PrintConsole(hOutput,
				             L"CloseHandle() returned error %lu\n",
				             GetLastError());
		}
	}

	// https://msdn.microsoft.com/en-us/library/ms682659.aspx

	ExitThread(dwError);
}
<<NOKEEP

AMD64-CRT.OBJ: AMD64-MSC.MAK
	CL.EXE /Brepro /c /Fo$@ /GAFwy /J /nologo /O2txy /Tc<< /W4 /Zl
// Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

// https://msdn.microsoft.com/en-us/library/7f0aews7.aspx

#pragma comment(compiler)
#pragma comment(lib, "kernel32.lib")
#pragma comment(user, "(C)opyright 2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>")

#define STRICT
#define WIN32_LEAN_AND_MEAN

#include <windows.h>

// https://msdn.microsoft.com/en-us/library/3ze4ytsc.aspx
// https://msdn.microsoft.com/en-us/library/6wd819wh.aspx
// https://msdn.microsoft.com/en-us/library/bky3b5dh.aspx

extern	int	main(int argc, char const *argv[], char const *envp[]);
extern	int	wmain(int argc, wchar_t const *argv[], wchar_t const *envp[]);

// https://msdn.microsoft.com/en-us/library/f9t8842e.aspx

__declspec(noreturn)
__declspec(safebuffers)
VOID	CDECL	mainCRTStartup(VOID)
{
	LPSTR	lpArgument;

	// https://msdn.microsoft.com/en-us/library/ms683187.aspx

	LPCSTR	lpBlock = GetEnvironmentStringsA();

	// https://msdn.microsoft.com/en-us/library/ms683156.aspx

	LPCSTR	lpCmdLine = GetCommandLineA();

	LPCSTR	lpCount = lpCmdLine;
	UINT	uiCount = 0U;
	UINT	uiQuote = 0U;
	UINT	argc;
	LPCSTR	*argv;
	LPCSTR	*envp;

	argc = (*lpCount != ' ' && *lpCount != '\t');

	while (*lpCount != '\0')	// count arguments
	{
		if (!uiQuote		// whitespace outside double quotes?
		 && (*lpCount == ' ' || *lpCount == '\t'))
		{
			do		// skip unquoted whitespace
				lpCount++;
			while (*lpCount == ' ' || *lpCount == '\t');

			argc += (*lpCount != '\0');

			uiCount = 0U;

			continue;
		}
		else if (*lpCount == '\\')
			uiCount ^= ~0U;
		else if (!uiCount	// unescaped double quote?
		      && *lpCount == '"')
			uiQuote ^= ~0U;
		else			// regular character
			uiCount = 0U;

		lpCount++;
	}

	if (uiQuote)			// unpaired double quote?

		// https://msdn.microsoft.com/en-us/library/ms680627.aspx

		SetLastError(ERROR_BAD_ARGUMENTS);

	// https://msdn.microsoft.com/en-us/library/wb1s57t5.aspx

	argv = (LPCSTR *) _alloca((argc + 1U) * sizeof(*argv) + (lpCount + 1U - lpCmdLine) * sizeof(**argv));

	argv[0] = lpArgument = (LPSTR) (argv + argc + 1U);

	argc = uiCount = uiQuote = 0U;

	while (*lpCmdLine != '\0')	// process arguments
	{
		if (!uiQuote		// whitespace outside double quotes?
		 && (*lpCmdLine == ' ' || *lpCmdLine == '\t'))
		{			// terminate current argument
			*lpArgument = '\0';

			do		// skip unquoted whitespace
				lpCmdLine++;
			while (*lpCmdLine == ' ' || *lpCmdLine == '\t');

			if (*lpCmdLine != '\0')
					// store address of next argument
				argv[++argc] = lpArgument = (LPSTR) lpCmdLine;

			uiCount = 0U;
		}
		else if (*lpCmdLine == '\\')
		{
			*lpArgument++ = *lpCmdLine++;
			uiCount++;	// count backslash
		}
		else if (*lpCmdLine == '"')
		{
			lpArgument -= (uiCount + 1U) / 2U;

			if (uiCount & 1U)
					// double quote preceeded by odd number
					//  of backslashes: keep half of them
					//   and the (escaped) double quote
				*lpArgument++ = *lpCmdLine++;
			else		// double quote preceeded by even number
					//  of backslashes: keep half of them
				if (*++lpCmdLine == '"' && uiQuote)
					// double quote inside double quotes and
					//  followed by another double quote:
					//   keep one double quote
					*lpArgument++ = *lpCmdLine++;
				else	// skip double quote and toggle state
					uiQuote ^= ~0U;
			uiCount = 0U;
		}
		else			// regular character
		{
			*lpArgument++ = *lpCmdLine++;
			uiCount = 0U;
		}
	}

	*lpArgument = '\0';		// terminate (last) argument

	argv[++argc] = NULL;		// store terminating NULL pointer
	envp = argv + argc;

	if (lpBlock != NULL)
	{
		for (uiCount = 0U,	// count environment strings
		     lpCount = lpBlock; *lpCount != '\0'; lpCount += strlen(lpCount) + 1U)
			if (*lpCount != '=')
				uiCount++;

		if (uiCount > 0U)	// process environment strings
		{
			// https://msdn.microsoft.com/en-us/library/wb1s57t5.aspx

			envp = (LPCSTR *) _alloca((uiCount + 1U) * sizeof(*envp));

			for (uiCount = 0U,
			     lpCount = lpBlock; *lpCount != '\0'; lpCount += strlen(lpCount) + 1U)
				if (*lpCount != '=')
					envp[uiCount++] = lpCount;

			envp[uiCount] = (LPCSTR) NULL;
		}
	}

	// https://msdn.microsoft.com/en-us/library/ms686036.aspx
	// https://msdn.microsoft.com/en-us/library/dd318070.aspx

	if (!SetConsoleOutputCP(GetACP()))
		/* no error handling */;

	// https://msdn.microsoft.com/en-us/library/ms682658.aspx

	ExitProcess(main(argc, argv, envp));
}

// https://msdn.microsoft.com/en-us/library/f9t8842e.aspx

__declspec(noreturn)
__declspec(safebuffers)
VOID	CDECL	wmainCRTStartup(VOID)
{
	LPWSTR	lpArgument;

	// https://msdn.microsoft.com/en-us/library/ms683187.aspx

	LPCWSTR	lpBlock = GetEnvironmentStringsW();

	// https://msdn.microsoft.com/en-us/library/ms683156.aspx

	LPCWSTR	lpCmdLine = GetCommandLineW();

	LPCWSTR	lpCount = lpCmdLine;
	UINT	uiCount = 0U;
	UINT	uiQuote = 0U;
	UINT	argc;
	LPCWSTR	*argv;
	LPCWSTR	*envp;

	argc = (*lpCount != L' ' && *lpCount != L'\t');

	while (*lpCount != L'\0')	// count arguments
	{
		if (!uiQuote		// whitespace outside double quotes?
		 && (*lpCount == L' ' || *lpCount == L'\t'))
		{
			do		// skip unquoted whitespace
				lpCount++;
			while (*lpCount == L' ' || *lpCount == L'\t');

			argc += (*lpCount != L'\0');

			uiCount = 0U;

			continue;
		}
		else if (*lpCount == L'\\')
			uiCount ^= ~0U;
		else if (!uiCount	// unescaped double quote?
		      && *lpCount == L'"')
			uiQuote ^= ~0U;
		else			// regular character
			uiCount = 0U;

		lpCount++;
	}

	if (uiQuote)			// unpaired double quote?

		// https://msdn.microsoft.com/en-us/library/ms680627.aspx

		SetLastError(ERROR_BAD_ARGUMENTS);

	// https://msdn.microsoft.com/en-us/library/wb1s57t5.aspx

	argv = (LPCWSTR *) _alloca((argc + 1U) * sizeof(*argv) + (lpCount + 1U - lpCmdLine) * sizeof(**argv));

	argv[0] = lpArgument = (LPWSTR) (argv + argc + 1U);

	argc = uiCount = uiQuote = 0U;

	while (*lpCmdLine != L'\0')	// process arguments
	{
		if (!uiQuote		// whitespace outside double quotes?
		 && (*lpCmdLine == L' ' || *lpCmdLine == L'\t'))
		{			// terminate current argument
			*lpArgument = L'\0';

			do		// skip unquoted whitespace
				lpCmdLine++;
			while (*lpCmdLine == L' ' || *lpCmdLine == L'\t');

			if (*lpCmdLine != L'\0')
					// store address of next argument
				argv[++argc] = lpArgument = (LPWSTR) lpCmdLine;

			uiCount = 0U;
		}
		else if (*lpCmdLine == L'\\')
		{
			*lpArgument++ = *lpCmdLine++;
			uiCount++;	// count backslash
		}
		else if (*lpCmdLine == L'"')
		{
			lpArgument -= (uiCount + 1U) / 2U;

			if (uiCount & 1U)
					// double quote preceeded by odd number
					//  of backslashes: keep half of them
					//   and the (escaped) double quote
				*lpArgument++ = *lpCmdLine++;
			else		// double quote preceeded by even number
					//  of backslashes: keep half of them
				if (*++lpCmdLine == L'"' && uiQuote)
					// double quote inside double quotes and
					//  followed by another double quote:
					//   keep one double quote
					*lpArgument++ = *lpCmdLine++;
				else	// skip double quote and toggle state
					uiQuote ^= ~0U;
			uiCount = 0U;
		}
		else			// regular character
		{
			*lpArgument++ = *lpCmdLine++;
			uiCount = 0U;
		}
	}

	*lpArgument = L'\0';		// terminate (last) argument

	argv[++argc] = NULL;		// store terminating NULL pointer
	envp = argv + argc;

	if (lpBlock != NULL)
	{
		for (uiCount = 0U,	// count environment strings
		     lpCount = lpBlock; *lpCount != L'\0'; lpCount += wcslen(lpCount) + 1U)
			if (*lpCount != L'=')
				uiCount++;

		if (uiCount > 0U)	// process environment strings
		{
			// https://msdn.microsoft.com/en-us/library/wb1s57t5.aspx

			envp = (LPCWSTR *) _alloca((uiCount + 1U) * sizeof(*envp));

			for (uiCount = 0U,
			     lpCount = lpBlock; *lpCount != L'\0'; lpCount += wcslen(lpCount) + 1U)
				if (*lpCount != L'=')
					envp[uiCount++] = lpCount;

			envp[uiCount] = (LPCWSTR) NULL;
		}
	}

	// https://msdn.microsoft.com/en-us/library/ms682658.aspx

	ExitProcess(wmain(argc, argv, envp));
}
<<NOKEEP

AMD64-FMT.OBJ: AMD64-MSC.MAK
	CL.EXE /Brepro /c /DOPTIMIZE /Fo$@ /GAFwy /J /nologo /O2txy /Tc<< /W4 /wd4100 /wd4213 /wd4244 /wd4554 /Zl
// Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

// https://msdn.microsoft.com/en-us/library/7f0aews7.aspx

#pragma comment(compiler)
#pragma comment(lib, "kernel32.lib")
#pragma comment(user, "(C)opyright 2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>")

#ifndef _WIN64
typedef	unsigned long	size_t;
#endif

typedef	unsigned short	wchar_t;

// https://msdn.microsoft.com/en-us/library/k64h826d.aspx

typedef	char	*va_list;

#ifdef _M_AMD64
// https://msdn.microsoft.com/en-us/library/0zz611kb.aspx

void	__va_start(va_list *, ...);
void	*__va_arg(va_list *, ...);
void	__va_end(va_list *);

#pragma intrinsic(__va_arg, __va_end, __va_start)

// https://msdn.microsoft.com/en-us/library/kb57fad8.aspx

#define va_arg(ptr, type)	((sizeof(type) > sizeof(long long) || (sizeof(type) & (sizeof(type) - 1)) != 0) ? **(type **) ((ptr += sizeof(long long)) - sizeof(long long)) : *(type *) ((ptr += sizeof(long long)) - sizeof(long long)))
#define va_copy(dst, src)	((dst) = (src))
#define va_end(ptr)		(ptr = (va_list) 0)
#define va_start(ptr, arg)	__va_start(&ptr, arg)
#elif _M_IX86
#define va_arg(ptr, type)	(*(type *) ((ptr += ((sizeof(type) + sizeof(int) - 1) & (0 - sizeof(int)))) - ((sizeof(type) + sizeof(int) - 1) & (0 - sizeof(int)))))
#define va_copy(dst, src)	((dst) = (src))
#define va_end(ptr)		(ptr = (va_list) 0)
#define va_start(ptr, arg)	(ptr = (va_list) &arg + ((sizeof(arg) + sizeof(int) - 1) & (0 - sizeof(int))))
#endif

#define	NULL	(void *) 0

// https://msdn.microsoft.com/en-us/library/dswaw1wk.aspx
// https://msdn.microsoft.com/en-us/library/hhta9830.aspx

#define memcpy	__movsb

// https://msdn.microsoft.com/en-us/library/1fdeehz6.aspx
// https://msdn.microsoft.com/en-us/library/ywwcatsa.aspx

#define memset	__stosb

// https://msdn.microsoft.com/en-us/library/dswaw1wk.aspx
// https://msdn.microsoft.com/en-us/library/adf7db3x.aspx

#define wmemcpy	__movsw

// https://msdn.microsoft.com/en-us/library/1fdeehz6.aspx
// https://msdn.microsoft.com/en-us/library/k99wy47b.aspx

#define wmemset	__stosw

// https://msdn.microsoft.com/en-us/library/ms776446.aspx
// https://msdn.microsoft.com/en-us/library/dd317756.aspx
// https://msdn.microsoft.com/en-us/library/ms776413.aspx
// https://msdn.microsoft.com/en-us/library/dd319072.aspx

#define CP_ACP			0U
#define MB_ERR_INVALID_CHARS	1UL
#define MB_PRECOMPOSED		8UL

__declspec(dllimport)
int	MultiByteToWideChar(unsigned int  uiCodePage,
	                    unsigned long dwFlags,
	                    char const    *lpMultiByte,
	                    int           cbMultiByte,
	                    wchar_t const *lpWideChar,
	                    int           cchWideChar);

__declspec(align(256))
static
const	char	ANSI[256] = "\0\1\2\3\4\5\6\a\b\t\n\v\f\r\16\17"
		            "\x10\x11\x12\x13\x14\x15\x16\x17\x18\x19\x1A\x1B\x1C\x1D\x1E\x1F"
		            "\x20\x21\x22\x23\x24\x25\x26\x27\x28\x29\x2A\x2B\x2C\x2D\x2E\x2F" // " !"#$$%&'()*+,-./"
		            "\x30\x31\x32\x33\x34\x35\x36\x37\x38\x39\x3A\x3B\x3C\x3D\x3E\x3F" // "0123456789:;<=>?"
		            "\x40\x41\x42\x43\x44\x45\x46\x47\x48\x49\x4A\x4B\x4C\x4D\x4E\x4F" // "@ABCDEFGHIJKLMNO"
		            "\x50\x51\x52\x53\x54\x55\x56\x57\x58\x59\x5A\x5B\x5C\x5D\x5E\x5F" // "PQRSTUVWXYZ[\\]^_"
		            "\x60\x61\x62\x63\x64\x65\x66\x67\x68\x69\x6A\x6B\x6C\x6D\x6E\x6F" // "`abcdefghijklmno"
		            "\x70\x71\x72\x73\x74\x75\x76\x77\x78\x79\x7A\x7B\x7C\x7D\x7E\x7F" // "pqrstuvwxyz{|}~ "
		            "\x80\x81\x82\x83\x84\x85\x86\x87\x88\x89\x8A\x8B\x8C\x8D\x8E\x8F"
		            "\x90\x91\x92\x93\x94\x95\x96\x97\x98\x99\x9A\x9B\x9C\x9D\x9E\x9F"
		            "\xA0\xA1\xA2\xA3\xA4\xA5\xA6\xA7\xA8\xA9\xAA\xAB\xAC\xAD\xAE\xAF"
		            "\xB0\xB1\xB2\xB3\xB4\xB5\xB6\xB7\xB8\xB9\xBA\xBB\xBC\xBD\xBE\xBF"
		            "\xC0\xC1\xC2\xC3\xC4\xC5\xC6\xC7\xC8\xC9\xCA\xCB\xCC\xCD\xCE\xCF"
		            "\xD0\xD1\xD2\xD3\xD4\xD5\xD6\xD7\xD8\xD9\xDA\xDB\xDC\xDD\xDE\xDF"
		            "\xE0\xE1\xE2\xE3\xE4\xE5\xE6\xE7\xE8\xE9\xEA\xEB\xEC\xED\xEE\xEF"
		            "\xF0\xF1\xF2\xF3\xF4\xF5\xF6\xF7\xF8\xF9\xFA\xFB\xFC\xFD\xFE\xFF";

__declspec(align(512))
static	wchar_t	UCS2[256];

__forceinline
int	tiebreak(char const *decimal, int start, int end)
{
	while (start < end)
		if (decimal[start++] > 0U)
			return 1;
	return 0;
}

// https://msdn.microsoft.com/en-us/library/56e442dc.aspx
// https://msdn.microsoft.com/en-us/library/8aky45ct.aspx
// https://msdn.microsoft.com/en-us/library/25366k66.aspx
// https://msdn.microsoft.com/en-us/library/0ecbz014.aspx

// '%' [{ ' ' | '#' | '\'' | '0' | '+' | '-' } ... ] [{ width | '*' }] [ '.' { precision | '*' }] [{ 'h' | 'hh' | 'l' | 'll' | 'I32' | 'I64' | 'L' | 't' | 'z' }] { 'a' | 'A' | 'c' | 'C' | 'd' | 'e' | 'E' | 'f' | 'F' | 'i' | 'n' | 'o' | 'p' | 's' | 'S' | 'u' | 'x' | 'X' }

// https://msdn.microsoft.com/en-us/library/tcxf1dw6.aspx

// signed int			d, i, n, o, x, X
// unsigned int			u
#ifdef C99 // also POSIX
// signed char		hh	d, i, n, o, x, X
// unsigned char	hh	u
#endif
// signed short		h	d, i, n, o, x, X
// unsigned short	h	u
// signed long		l	d, i, n, o, x, X
// unsigned long	l	u
#ifdef C99 // also POSIX
// signed long long	ll	d, i, n, o, x, X
// unsigned long long	ll	u
// unsigned void *	t	u
// signed void *	z	d, i, n, o, x, X
#endif

// https://msdn.microsoft.com/en-us/library/hf4y5e3w.aspx

//				d = signed integer to base 10, optionally prefixed [+- 0]
//				i = signed integer to base 10, optionally prefixed [+- 0]
//				o = unsigned integer to base 8, optionally prefixed [+-0]
//				u = unsigned integer to base 10, unsigned
//				x = unsigned integer to base 16, optionally prefixed 0x
//				X = unsigned integer to base 16, optionally prefixed 0X
//
// int				n = character count
// void *			p = unsigned pointer to base 16, optionally prefixed 0x
// void *			P = unsigned pointer to base 16, optionally prefixed 0X
//
// double			a = IEEE-754 double to base 16
// double			A = IEEE-754 double to base 16
//
// double			e = IEEE-754 double to decimal
// double			E = IEEE-754 double to decimal
//
// double		l	f, g, G			(NOT SUPPORTED)
// long double		ll	e, E, f, g, G		(NOT SUPPORTED)
// long double		L	e, E, f, g, G		(NOT SUPPORTED)
//
// char				C			(UNICODE function)
// char				c			(ANSI function)
// char			h	c, C			(ANSI & UNICODE function)
// wchar_t			C			(ANSI function)
// wchar_t			c			(UNICODE function)
// wchar_t		l	c, C			(ANSI & UNICODE function)
// char	*			S			(UNICODE function)
// char	*			s, [...]		(ANSI function)
// char	*		h	s, S			(ANSI & UNICODE function)
// wchar_t *			S			(ANSI function)
// wchar_t *			s, [...]		(UNICODE function)
// wchar_t *		l	s, S			(ANSI & UNICODE function)

// https://msdn.microsoft.com/en-us/library/1kt27hek.aspx

__declspec(safebuffers)
int	vsnprintf(char *output, size_t count, char const *format, va_list input)
{
	char	const * const start = format;
	char	const * const begin = output;
	char	const * const end = (count < 2) ? output : output + count - 2;
	char	const *ansi;
	wchar_t	const *unicode;
	wchar_t	cw;
	char	cb;
	char	pad;
	char	sign;
	char	size;
	char	scratch[24];		// characters from integer conversion:
					//  ~0ULL has 22 octal characters
	char	decimal[768];		// digits from floating-point conversion:
					//  2.0**-x has x fractional decimal digits;
					//  0x1p-1074 = 4.9406564584124654...e-324
					//   has 323 significant zeroes after the decimal point,
					//    followed by 751 digits;
					//  0x0.FFFFFFFFFFFFFp-1022 = 2.2250738585072008...e-308
					//   has 307 significant zeroes after the decimal point,
					//    followed by 767 digits;
					//  0x1p+1024 = 1.7976931348623159...e+308 has 309 digits
	char	digit;
	int	length;			// number of characters from conversion
	int	index;			// index of first digit from conversion
	int	left;
	int	hash;
	int	blank;
//	int	group;
	int	width;
	int	precision;
	int	exponent;
	int	point;
	int	extra;

	unsigned long		ul;
	unsigned long long	ull;

	if (UCS2['\t'] == L'\0')	// ANSI to UTF-16LE conversion table not initialized?

		// https://msdn.microsoft.com/en-us/library/ms776413.aspx
		// https://msdn.microsoft.com/en-us/library/dd319072.aspx

		if (MultiByteToWideChar(CP_ACP,
		                        MB_ERR_INVALID_CHARS | MB_PRECOMPOSED,
		                        ANSI, 256,
		                        UCS2, 256) != 256)
			return ~2147483647;

	for (;;)
	{
		while (*format != '%'	// copy format string up to next conversion specifier
		    && *format != '\0'	//  or its end to output string
		    && output < end)
			*output++ = *format++;

		if (*format == '\0')	// format string processed completely?
			break;

		if (*++format == '%')	// double '%' in format string?
		{
			if (output < end)
				*output++ = *format++;
			continue;
		}
					// conversion specifier in format string
		pad = ' ';
		sign = '\0';
		size = '\0';
		left = 0;
		hash = 0;
		blank = 0;
	//	group = 0;
		width = 0;
		length = 0;
		precision = -1;

		do			// flag(s) in format string?
			if (*format == '-')
				left = 1;
			else if (*format == '+')
				sign = '+';
			else if (*format == '0')
				pad = '0';
			else if (*format == ' ')
				blank = 1;
			else if (*format == '#')
				hash = 1;
		//	else if (*format == '\'')
		//		group = 1;
			else
				break;
		while (*++format != '\0');

		if (blank && sign == '\0')
			sign = ' ';

		if (*format >= '0'	// field width in format string?
		 && *format <= '9')
		{
			for (width = *format++ - '0';
			     *format >= '0' && *format <= '9';
			     width *= 10, width += *format++ - '0')
				continue;

			if (width < 0)
				width = 2147483647;
		}
		else if (*format == '*')
		{
			format++;

			width = va_arg(input, int);

			if (width < 0)
			{
				if (width == ~2147483647)
					width = 0;
				else
					width = -width;

				left = 1;
			}
		}

	//	if (left)
	//		pad = ' ';

		if (*format == '.')	// precision in format string?
		{
			pad = ' ';

			if (*++format == '*')
			{
				format++;

				precision = va_arg(input, int);

			//	if (precision < 0)
			//		precision = -1;
			}
			else if (*format >= '0' && *format <= '9')
			{
				for (precision = *format++ - '0';
				     *format >= '0' && *format <= '9';
				     precision *= 10, precision += *format++ - '0')
					continue;

				if (precision < 0)
					precision = 2147483647;
			}
			else
				precision = 0;
		}

		if (*format == 'j'	// size modifier in format string?
		 || *format == 't'
		 || *format == 'z')
			size = *format++;
		else if (*format == 'h'
		      || *format == 'l')
		{
			size = *format++;

			if (size == *format)
			{
				format++;
					// encode "hh" as 'H' and "ll" as 'L'
				size ^= 'a' ^ 'A';
			}
		}

		if (*format == 'd'	// conversion type in format string?
		 || *format == 'i')
		{			// signed decimal integer:
					//  [ ' ' ... | '+' | '-' ] [ '0' ... ] [ <digit> ... ] [ ' ' ... ]
			if (size == 'j'
			 || size == 't'
			 || size == 'z'
			 || size == 'L')
				(long long) ull = va_arg(input, long long);
			else if (size == 'h')
				(long long) ull = va_arg(input, short);
			else if (size == 'H')
				(long long) ull = va_arg(input, char);
			else // 'l', 't', 'z', '\0'
				(long long) ull = va_arg(input, int);

			if ((long long) ull < 0)
			{
				ull = 0 - ull;
				sign = '-';
			}
					// convert to decimal characters,
					//  least significant digit in scratch[0],
					//   most significant digit in scratch[length - 1]
			do
				scratch[length++] = '0' + (char) (ull % 10);
			while (ull /= 10);
					// index of decimal integer (leading '0' padding)
			index = length;

			if (precision > length)
				precision -= length;
			else
				precision = 0;

			if (sign != '\0')
				scratch[length++] = sign;

			goto NUMBER;
		}
		else if (*format == 'u')
		{			// unsigned decimal integer:
					//  [{ ' ' | '0' } ... ] [ <digit> ... ] [ ' ' ... ]
			if (size == 'j'
			 || size == 't'
			 || size == 'z'
			 || size == 'L')
				ull = va_arg(input, unsigned long long);
			else if (size == 'h')
				ull = va_arg(input, unsigned short);
			else if (size == 'H')
				ull = va_arg(input, unsigned char);
			else // 'l', 't', 'z', '\0'
				ull = va_arg(input, unsigned int);

					// convert to decimal characters,
					//  least significant digit in scratch[0],
					//   most significant digit in scratch[length - 1]
			do
				scratch[length++] = '0' + (char) (ull % 10);
			while (ull /= 10);
					// index of decimal integer (leading '0' padding)
			index = length;

			if (precision > length)
				precision -= length;
			else
				precision = 0;

			goto NUMBER;
		}
		else if (*format == 'o')
		{			// unsigned octal integer:
					//  [{ ' ' | '0' } ... ] [ <digit> ... ] [ ' ' ... ]
			if (size == 'j'
			 || size == 't'
			 || size == 'z'
			 || size == 'L')
				ull = va_arg(input, unsigned long long);
			else if (size == 'h')
				ull = va_arg(input, unsigned short);
			else if (size == 'H')
				ull = va_arg(input, unsigned char);
			else // 'l', 't', 'z', '\0'
				ull = va_arg(input, unsigned int);

			if (ull == 0)
				hash = 0;
			do
				scratch[length++] = '0' + (char) (ull % 8);
			while (ull /= 8);
					// index of octal integer (leading '0' padding)
			index = length;

			if (hash)
#if 0
				if (precision >= 0)
					precision++;
#else
				precision += (precision >= 0);
#endif
			if (precision > length)
				precision -= length;
			else
				precision = 0;
					// prefix '0'
			if (hash)
				scratch[length++] = '0';

			goto NUMBER;
		}
#if 0
		else if (*format == 'x'
		      || *format == 'X')
#else
		else if ((*format | 'x' ^ 'X') == 'x')
#endif
		{			// unsigned hexadecimal integer:
					//  [ ' ' ... ] [ '0' { 'x' | 'X' }] [ '0' ... ] [ <digit> ... ] [ ' ' ... ]

			ansi = "0123456789ABCDEF0123456789abcdef" + 16 * (*format ^ 'X') / ('x' ^ 'X');

			if (size == 'j'
			 || size == 't'
			 || size == 'z'
			 || size == 'L')
				ull = va_arg(input, unsigned long long);
			else if (size == 'h')
				ull = va_arg(input, unsigned short);
			else if (size == 'H')
				ull = va_arg(input, unsigned char);
			else // 'l', 't', 'z', '\0'
				ull = va_arg(input, unsigned int);

			if (ull == 0)
				hash = 0;
			do
				scratch[length++] = ansi[ull % 16];
			while (ull /= 16);
					// index of hexadecimal integer (leading '0' padding)
			index = length;

			if (precision > length)
				precision -= length;
			else
				precision = 0;

			if (hash)
			{		// prefix "0x" or "0X"
				scratch[length++] = *format;
				scratch[length++] = '0';
			}

		NUMBER:	if (width > precision + length)
				width -= precision + length;
			else
				width = 0;

			if (left)
				while (length > index && output < end)
					*output++ = scratch[--length];
			else if (pad == '0')
			{
				while (length > index && output < end)
					*output++ = scratch[--length];
#ifndef OPTIMIZE
				while (width-- > 0 && output < end)
					*output++ = '0';
#else
				if (width > end - output)
					width = end - output;

				// https://msdn.microsoft.com/en-us/library/1fdeehz6.aspx

				memset(output, '0', width);

				output += width;
#endif // OPTIMIZE
			}
			else
			{
#ifndef OPTIMIZE
				while (width-- > 0 && output < end)
					*output++ = ' ';
#else
				if (width > end - output)
					width = end - output;

				// https://msdn.microsoft.com/en-us/library/1fdeehz6.aspx

				memset(output, ' ', width);

				output += width;
#endif // OPTIMIZE
				while (length > index && output < end)
					*output++ = scratch[--length];
			}
#ifndef OPTIMIZE
			while (precision-- > 0 && output < end)
				*output++ = '0';
#else
			if (precision > end - output)
				precision = end - output;

			// https://msdn.microsoft.com/en-us/library/1fdeehz6.aspx

			memset(output, '0', precision);

			output += precision;
#endif // OPTIMIZE
			while (length > 0 && output < end)
				*output++ = scratch[--length];
#ifndef OPTIMIZE
		//	if (left)
		RIGHT:		while (width-- > 0 && output < end)
					*output++ = ' ';
#else
		RIGHT:	if (left)
			{
				if (width > end - output)
					width = end - output;

				// https://msdn.microsoft.com/en-us/library/1fdeehz6.aspx

				memset(output, ' ', width);

				output += width;
			}
#endif // OPTIMIZE
		}
#if 0
		else if (*format == 'p'
		      || *format == 'P')
#else
		else if ((*format | 'p' ^ 'P') == 'p')
#endif
		{			// unsigned hexadecimal pointer:
					//  [ ' ' ... ] [ '0' { 'x' | 'X' }] [ '0' ... ] [ <digit> ... ] [ ' ' ... ]
			if (size != '\0'
			 || sign != '\0'
			 || pad != ' '
			 || precision != -1)
				goto INVALID;

			ansi = "0123456789ABCDEF0123456789abcdef" + 16 * (*format ^ 'P') / ('p' ^ 'P');

			ull = va_arg(input, unsigned long long);

			if (ull == 0)
				hash = 0;
			do
				scratch[length++] = ansi[ull % 16];
			while (ull /= 16);

			while (length < 2 * sizeof(ull))
				scratch[length++] = '0';

			if (hash)
			{		// prefix "0x" or "0X"
				scratch[length++] = *format ^ 'P' | 'X';
				scratch[length++] = '0';
			}

			if (!left)
				while (width-- > length && output < end)
					*output++ = ' ';

			while (length > 0 && output < end)
				*output++ = scratch[--length];

			goto RIGHT;
		}
		else if (*format == 'c')
		{			// ANSI or UNICODE character:
					//  [ ' ' ... ] [ <character> ] [ ' ' ... ]
			if (precision != -1)
				goto INVALID;

			if (size == 'l')
				goto C;	// UNICODE character?

			if (size != '\0'
			 && size != 'h')
				goto INVALID;

		c:	cb = va_arg(input, char);

		COMMON:	if (cb != '\0')
#if 0
				if (width > 0)
					width--;
#else
				width -= (width > 0);
#endif
#ifndef OPTIMIZE
			if (!left)
				while (width-- > 0 && output < end)
					*output++ = ' ';
#else
			if (!left)
			{
				if (width > end - output)
					width = end - output;

				// https://msdn.microsoft.com/en-us/library/1fdeehz6.aspx

				memset(output, ' ', width);

				output += width;
			}
#endif // OPTIMIZE
			if (cb != '\0' && output < end)
				*output++ = cb;

			goto RIGHT;
		}
		else if (*format == 'C')
		{			// ANSI or UNICODE character:
					//  [ ' ' ... ] [ <character> ] [ ' ' ... ]
			if (precision != -1)
				goto INVALID;

			if (size == 'h')
				goto c;	// ANSI character?

			if (size != '\0'
			 && size != 'l')
				goto INVALID;

		C:	cw = va_arg(input, wchar_t);

			if (cw == UCS2[cw & L'\xFF'])
				cb = (char) cw;
			else
				for (cb = '?', ul = 0x80; ul < 0xA0; ul++)
					if (cw == UCS2[ul])
					{
						cb = (char) ul;
						break;
					}
			goto COMMON;
		}
		else if (*format == 's')
		{			// ANSI or UNICODE character string:
					//  [ ' ' ... ] [ <character> ... ] [ ' ' ... ]
			if (size == 'l')
				goto S;	// UNICODE character string?

			if (size != '\0'
			 && size != 'h')
				goto INVALID;

		s:	ansi = va_arg(input, char *);

			if (ansi == NULL)
				ansi = "(null)";
#if 0
			// https://msdn.microsoft.com/en-us/library/78zh94ax.aspx

			length = strlen(ansi);

			if (precision >= 0 && length > precision)
				length = precision;
#else
			if (precision < 0)

				// https://msdn.microsoft.com/en-us/library/78zh94ax.aspx

				length = strlen(ansi);
			else
				for (length = 0; length < precision; length++)
					if (ansi[length] == '\0')
						break;
#endif
		STRING:	if (width > length)
				width -= length;
			else
				width = 0;
#ifndef OPTIMIZE
			if (!left)
				while (width-- > 0 && output < end)
					*output++ = ' ';

			while (length-- > 0 && output < end)
				*output++ = *ansi++;
#else
			if (!left)
			{
				if (width > end - output)
					width = end - output;

				// https://msdn.microsoft.com/en-us/library/1fdeehz6.aspx

				memset(output, ' ', width);

				output += width;
			}

			if (length > end - output)
				length = end - output;

			// https://msdn.microsoft.com/en-us/library/dswaw1wk.aspx

			memcpy(output, ansi, length);

			output += length;
#endif // OPTIMIZE
			goto RIGHT;
		}
		else if (*format == 'S')
		{			// ANSI or UNICODE character string:
					//  [ ' ' ... ] [ <character> ... ] [ ' ' ... ]
			if (size == 'h')
				goto s;	// ANSI character string?

			if (size != '\0'
			 && size != 'l')
				goto INVALID;

		S:	unicode = va_arg(input, wchar_t *);

			if (unicode == NULL)
				unicode = L"(null)";
#if 0
			// https://msdn.microsoft.com/en-us/library/78zh94ax.aspx

			length = wcslen(unicode);

			if (precision >= 0 && length > precision)
				length = precision;
#else
			if (precision < 0)

				// https://msdn.microsoft.com/en-us/library/78zh94ax.aspx

				length = wcslen(unicode);
			else
				for (length = 0; length < precision; length++)
					if (unicode[length] == L'\0')
						break;
#endif
			if (width > length)
				width -= length;
			else
				width = 0;
#ifndef OPTIMIZE
			if (!left)
				while (width-- > 0 && output < end)
					*output++ = ' ';
#else
			if (!left)
			{
				if (width > end - output)
					width = end - output;

				// https://msdn.microsoft.com/en-us/library/1fdeehz6.aspx

				memset(output, ' ', width);

				output += width;
			}
#endif // OPTIMIZE
			while (length-- > 0 && output < end)
			{
				cw = *unicode++;

				if (cw == UCS2[cw & L'\xFF'])
					cb = (char) cw;
				else
					for (cb = '?', ul = 0x80; ul < 0xA0; ul++)
						if (cw == UCS2[ul])
						{
							cb = (char) ul;
							break;
						}

				*output++ = cb;
			}

			goto RIGHT;
		}
		else if (*format == 'n')
		{			// assign character count
			if (size == 'j'
			 || size == 't'
			 || size == 'z'
			 || size == 'L')
				*va_arg(input, long long *) = (long long) (output - begin);
			else if (size == 'h')
				*va_arg(input, short *) = (short) (output - begin);
			else if (size == 'H')
				*va_arg(input, char *) = (char) (output - begin);
			else
				*va_arg(input, int *) = (int) (output - begin);
		}
#if 0
		else if (*format == 'a'
		      || *format == 'A')
#else
		else if ((*format | 'a' ^ 'A') == 'a')
#endif
		{			// hexadecimal double-precision floating-point:
					//  [ ' ' ... | '+' | '-' ] '0' { 'x' | 'X' } [ '0' ... ] <digit> [ '.' <digit> ... ] { 'p' | 'P' } { '+' | '-' } <digit> ... [ ' ' ... ]
			if (size != '\0')
				goto INVALID;

			ansi = "0123456789ABCDEF0123456789abcdef" + 16 * (*format ^ 'A') / ('a' ^ 'A');
			ull = va_arg(input, unsigned long long);

			if ((long long) ull < 0)
			{		// strip sign bit
				ull &= ~0ULL >> 1;
				sign = '-';
			}
					// biased exponent
			exponent = (int) (ull >> 52);

			if (exponent == 2047)
			{
				if (sign != '\0')
					scratch[length++] = sign;
#ifndef OPTIMIZE
				if (ull & ~0ULL >> 12)
				{
					scratch[length++] = *format ^ 'A' | 'N';
					scratch[length++] = *format ^ 'A' | 'A';
					scratch[length++] = *format ^ 'A' | 'N';
				}
				else
				{
					scratch[length++] = *format ^ 'A' | 'I';
					scratch[length++] = *format ^ 'A' | 'N';
					scratch[length++] = *format ^ 'A' | 'F';
				}

				scratch[length] = '\0';
#else
				// https://msdn.microsoft.com/en-us/library/dswaw1wk.aspx

				if (ull & ~0ULL >> 12)
					memcpy(scratch + length, *format == 'a' ? "nan" : "NAN", sizeof("NAN"));
				else
					memcpy(scratch + length, *format == 'a' ? "inf" : "INF", sizeof("INF"));

				length += 3;
#endif // OPTIMIZE
				ansi = scratch;

				goto STRING;
			}
			else if (exponent > 0)
			{		// unbiased exponent
				exponent -= 1023;
					// strip exponent, insert hidden bit
				ull &= ~0ULL >> 12;
				ull |= 1ULL << 52;
			}
			else		// 0.0 or denormal floating-point number
#ifndef OPTIMIZE
				if (ull)
				{	// unbiased exponent
					exponent = 1 - 1023;
					// normalize significand
					while (ull < 1ULL << 52)
						ull += ull, exponent--;
				}
#else
				// https://msdn.microsoft.com/en-us/library/fbxyd7zd.aspx

				if (_BitScanReverse64(&ul, ull))
				{	// unbiased exponent
					exponent = 1 - 1023 - (52 - ul);
					// normalize significand
					ull <<= 52 - ul;
				}
#endif // OPTIMIZE
					// convert significand to hexadecimal, one digit per byte,
					//  least significant digit in decimal[length - 1],
					//   most significant digit in decimal[0],
					//    hexadecimal point after most significant digit
			do
				decimal[length++] = (char) (ull >> 52);
			while (ull &= ~0ULL >> 12, ull <<= 4);

			point = 0;
			extra = 0;

			if (precision < 0)
				precision = length - 1;
			else if (precision > length - 1)
			{		// number of additional trailing zeroes
				extra = precision - length + 1;
				precision = length - 1;
				hash = 1;
			}
			else if (precision < length - 1
			      && (decimal[precision + 1] > 8U
			       || decimal[precision + 1] == 8U && decimal[precision] % 2U
			       || tiebreak(decimal, precision + 2, length)))
			{		// round to nearest, ties to even
#if 0
				index = precision + 1;
				digit = 8U;

				do
					digit += decimal[index],
					decimal[index] = digit % 16U;
				while (digit /= 16U && index-- > 0);

				if (digit)
				{	// hexadecimal overflow: increment exponent
					exponent++;
					// set leading digit
					decimal[0] = digit;
				}
#else
				index = precision;

				while (index >= 0 && decimal[index] == 15U)
					decimal[index--] = 0U;

				if (index < 0)
				{	// hexadecimal overflow: increment exponent
					exponent++;
					// set leading digit
					decimal[0] = 1U;
				}
				else
					decimal[index] += 1U;
#endif
			}
					// hexadecimal exponent
			if (exponent < 0)
			{
				exponent = -exponent;
				size = '-';
			}
			else if (exponent > 0)
				size = '+';
		//	else
		//		size = '\0';

			length = 0;

			do		// hexadecimal exponent
				scratch[length++] = '0' + (char) (exponent % 10U);
			while (exponent /= 10U);
					// exponent sign
			if (size != '\0')
				scratch[length++] = size;
					// exponent prefix 'p' or 'P'
			scratch[length++] = *format ^ 'A' | 'P';
					// index of exponent prefix (trailing '0' padding)
			exponent = length;
					// hexadecimal point
			if (precision > 0 || hash > 0)
				scratch[length++] = '.';
					// most significant digit
			scratch[length++] = '0' + decimal[0];
					// index of most significant digit (leading '0' padding)
			index = length;
					// prefix "0x" or "0X"
			scratch[length++] = *format ^ 'A' | 'X';
			scratch[length++] = '0';
					// sign
			if (sign != '\0')
				scratch[length++] = sign;

			goto DOUBLE;
		}
#if 0
		else if (*format == 'e'
		      || *format == 'E')
#else
		else if ((*format | 'e' ^ 'E') == 'e')
#endif
		{			// decimal double-precision floating-point:
					//  [ ' ' ... | '+' | '-' ] [ '0' ... ] <digit> [ '.' <digit> ... [ '0' ... ]] { 'e' | 'E' } [ '+' | '-' ] <digit> ... [ ' ' ... ]
			if (size != '\0')
				goto INVALID;

			ansi = NULL;

			ull = va_arg(input, unsigned long long);

			if ((long long) ull < 0)
			{		// strip sign bit
				ull &= ~0ULL >> 1;
				sign = '-';
			}
					// biased exponent
			exponent = (int) (ull >> 52);

			if (exponent == 2047)
			{
				if (sign != '\0')
					scratch[length++] = sign;
#ifndef OPTIMIZE
				if (ull & ~0ULL >> 12)
				{
					scratch[length++] = *format ^ 'E' | 'N';
					scratch[length++] = *format ^ 'E' | 'A';
					scratch[length++] = *format ^ 'E' | 'N';
				}
				else
				{
					scratch[length++] = *format ^ 'E' | 'I';
					scratch[length++] = *format ^ 'E' | 'N';
					scratch[length++] = *format ^ 'E' | 'F';
				}

				scratch[length] = '\0';
#else
				// https://msdn.microsoft.com/en-us/library/dswaw1wk.aspx

				if (ull & ~0ULL >> 12)
					memcpy(scratch + length, *format == 'e' ? "nan" : "NAN", sizeof("NAN"));
				else
					memcpy(scratch + length, *format == 'e' ? "inf" : "INF", sizeof("INF"));

				length += 3;
#endif // OPTIMIZE
				ansi = scratch;

				goto STRING;
			}
			else if (exponent > 0)
			{		// unbiased exponent
				exponent -= 1023;
					// strip exponent, insert hidden bit
				ull &= ~0ULL >> 12;
				ull |= 1ULL << 52;
			}
			else		// 0.0 or denormal floating-point number
#ifndef OPTIMIZE
				if (ull)
				{	// unbiased exponent
					exponent = 1 - 1023;
					// normalize significand
					while (ull < 1ULL << 52)
						ull += ull, exponent--;
				}
#else
				// https://msdn.microsoft.com/en-us/library/fbxyd7zd.aspx

				if (_BitScanReverse64(&ul, ull))
				{	// unbiased exponent
					exponent = 1 - 1023 - (52 - ul);
					// normalize significand
					ull <<= 52 - ul;
				}
#endif // OPTIMIZE
					// convert significand to decimal, one digit per byte,
					//  least significant digit in decimal[length - 1],
					//   most significant digit in decimal[0],
					//    decimal point after most significant digit:
					//     0x0                 = 0[.]
					//     0x1[.]0000000000000 = 1[.]
					//     0x1[.]0000000000001 = 1[.]0000000000000002220446049250313080847263336181640625
					//     0x1[.]8000000000000 = 1[.]5
					//     0x1[.]FFFFFFFFFFFFF = 1[.]9999999999999997779553950749686919152736663818359375
			do
				decimal[length++] = (char) (ull >> 52);
			while (ull &= ~0ULL >> 12, ull *= 10);

					// decimal point past unhidden bit
					//  (also decimal exponent)
			point = 0;
					// multiply by 2**exponent
			while (exponent > 0)
			{
				index = length;
				digit = 0U;

				if (decimal[0] < 5U)
				{
					do
						digit += decimal[index - 1] * 2U,
						decimal[index - 1] = digit % 10U, digit /= 10U;
					while (--index > 0);
				}
				else	// decimal overflow: move decimal point and all digits
				{	//  one place right, insert new non-zero leading digit
					point++;
					length++;

					do
						digit += decimal[index - 1] * 2U,
						decimal[index] = digit % 10U, digit /= 10U;
					while (--index > 0);

					decimal[0] = digit;
				}

				exponent--;
			}
					// divide by 2**-exponent
			while (exponent < 0)
			{
				index = 0;

				if (decimal[length - 1] % 2U)
					decimal[length++] = 0U;

				if (decimal[0] > 1U)
				{
					digit = 0U;

					do
						digit = digit * 10U + decimal[index],
						decimal[index] = digit / 2U, digit %= 2U;
					while (++index < length);
				}
				else	// decimal underflow: discard new leading 0, move
				{	//  decimal point and all digits one place left
					point--;
					length--;
					digit = decimal[0];

					do
						digit = digit * 10U + decimal[index + 1],
						decimal[index] = digit / 2U, digit %= 2U;
					while (++index < length);
				}

				exponent++;
			}

			if (precision < 0)
				precision = 6;

			extra = 0;

			if (precision > length - 1)
			{		// number of additional trailing zeroes
				extra = precision - length + 1;
				precision = length - 1;
				hash = 1;
			}
			else if (precision < length - 1
			      && (decimal[precision + 1] > 5U
			       || decimal[precision + 1] == 5U && decimal[precision] % 2U
			       || tiebreak(decimal, precision + 2, length)))
			{		// round to nearest, ties to even
#if 0
				index = precision + 1;
				digit = 5U;

				do
					digit += decimal[index],
					decimal[index] = digit % 10U;
				while (digit /= 10U && index-- > 0);

				if (digit)
				{	// decimal overflow: move decimal point right
					point++;
					// set leading digit
					decimal[0] = digit;
				}
#else
				index = precision;

				while (index >= 0 && decimal[index] == 9U)
					decimal[index--] = 0U;

				if (index < 0)
				{	// decimal overflow: move decimal point right
					point++;
					// set leading digit
					decimal[0] = 1U;
				}
				else
					decimal[index] += 1U;
#endif
			}
					// decimal exponent
			if (point < 0)
			{
				point = -point;
				size = '-';
			}
			else if (point > 0)
				size = '+';
		//	else
		//		size = '\0';

			do		// decimal exponent
				scratch[exponent++] = '0' + (char) (point % 10U);
			while (point /= 10U);
					// exponent sign
			if (size != '\0')
				scratch[exponent++] = size;
					// exponent prefix 'e' or 'E'
			scratch[exponent++] = *format;
					// index of exponent prefix (trailing '0' padding)
			length = exponent;
					// decimal point
			if (precision > 0 || hash > 0)
				scratch[length++] = '.';
					// most significant digit
			scratch[length++] = '0' + decimal[0];
					// index of most significant digit (leading '0' padding)
			index = length;
					// sign
			if (sign != '\0')
				scratch[length++] = sign;

		DOUBLE:	if (width > length + precision + extra)
				width -= length + precision + extra;
			else
				width = 0;

			if (left)
			{		// sign, prefix, most significant digit and point
				while (length > exponent && output < end)
					*output++ = scratch[--length];

				if (ansi == NULL)
					while (point < precision && output < end)
						*output++ = '0' + decimal[++point];
				else
					while (point < precision && output < end)
						*output++ = ansi[decimal[++point]];
#ifndef OPTIMIZE
				while (extra-- > 0 && output < end)
					*output++ = '0';
#else
				if (extra > end - output)
					extra = end - output;

				// https://msdn.microsoft.com/en-us/library/1fdeehz6.aspx

				memset(output, '0', extra);

				output += extra;
#endif // OPTIMIZE
				while (length > 0 && output < end)
					*output++ = scratch[--length];

				goto RIGHT;
			}

			if (pad == '0')
			{		// sign and prefix
				while (length > index && output < end)
					*output++ = scratch[--length];
#ifndef OPTIMIZE
				while (width-- > 0 && output < end)
					*output++ = '0';
#else
				if (width > end - output)
					width = end - output;

				// https://msdn.microsoft.com/en-us/library/1fdeehz6.aspx

				memset(output, '0', width);

				output += width;
#endif // OPTIMIZE
			}
			else
			{
#ifndef OPTIMIZE
				while (width-- > 0 && output < end)
					*output++ = ' ';
#else
				if (width > end - output)
					width = end - output;

				// https://msdn.microsoft.com/en-us/library/1fdeehz6.aspx

				memset(output, ' ', width);

				output += width;
#endif // OPTIMIZE
				while (length > index && output < end)
					*output++ = scratch[--length];
			}

			while (length > exponent && output < end)
				*output++ = scratch[--length];

			if (ansi == NULL)
				while (point < precision && output < end)
					*output++ = '0' + decimal[++point];
			else
				while (point < precision && output < end)
					*output++ = ansi[decimal[++point]];
#ifndef OPTIMIZE
			while (extra-- > 0 && output < end)
				*output++ = '0';
#else
			if (extra > end - output)
				extra = end - output;

			// https://msdn.microsoft.com/en-us/library/1fdeehz6.aspx

			memset(output, '0', extra);

			output += extra;
#endif // OPTIMIZE
			while (length > 0 && output < end)
				*output++ = scratch[--length];
		}
		else			// return negative index of invalid character
		INVALID:		//  in format string
			return (int) (start - format);

		format++;
	}

	if (output != NULL)
		*output = '\0';

	return (int) (output - begin);	// return length of output string
}

// https://msdn.microsoft.com/en-us/library/2ts7cx93.aspx

int	snprintf(char *output, size_t count, const char *format, ...)
{
	int	result;

	// https://msdn.microsoft.com/en-us/library/kb57fad8.aspx

	va_list	input;
	va_start(input, format);

	// https://msdn.microsoft.com/en-us/library/1kt27hek.aspx

	result = vsnprintf(output, count, format, input);

	va_end(input);

	return result;
}

// https://msdn.microsoft.com/en-us/library/28d5ce15.aspx

__declspec(safebuffers)
int	vswprintf(wchar_t *output, size_t count, wchar_t const *format, va_list input)
{
	wchar_t	const * const start = format;
	wchar_t	const * const begin = output;
	wchar_t	const * const end = (count < 2) ? output : output + count - 2;
	wchar_t	const *unicode;
	char	const *ansi;
	char	decimal[768];		// digits from floating-point conversion:
					//  2.0**-x has x fractional decimal digits;
					//  0x1p-1074 = 4.9406564584124654...e-324
					//   has 323 significant zeroes after the decimal point,
					//    followed by 751 digits;
					//  0x0.FFFFFFFFFFFFFp-1022 = 2.2250738585072008...e-308
					//   has 307 significant zeroes after the decimal point,
					//    followed by 767 digits;
					//  0x1p+1024 = 1.7976931348623159...e+308 has 309 digits
	char	digit;
	char	cb;
	wchar_t	cw;
	wchar_t	pad;
	wchar_t	sign;
	wchar_t	size;
	wchar_t	scratch[24];		// characters from integer conversion:
					//  ~0ULL has 22 octal characters
	int	length;			// number of characters from conversion
	int	index;			// index of first digit from conversion
	int	left;
	int	hash;
	int	blank;
//	int	group;
	int	width;
	int	precision;
	int	exponent;
	int	point;
	int	extra;

	unsigned long		ul;
	unsigned long long	ull;

	if (UCS2['\t'] == L'\0')	// ANSI to UTF-16LE conversion table not initialized?

		// https://msdn.microsoft.com/en-us/library/ms776413.aspx
		// https://msdn.microsoft.com/en-us/library/dd319072.aspx

		if (MultiByteToWideChar(CP_ACP,
		                        MB_ERR_INVALID_CHARS | MB_PRECOMPOSED,
		                        ANSI, 256,
		                        UCS2, 256) != 256)
			return ~2147483647;

	for (;;)
	{
		while (*format != L'%'	// copy format string up to next conversion specifier
		    && *format != L'\0'	//  or its end to output string
		    && output < end)
			*output++ = *format++;

		if (*format == L'\0')	// format string processed completely?
			break;

		if (*++format == L'%')	// double L'%' in format string?
		{
			if (output < end)
				*output++ = *format++;
			continue;
		}
					// conversion specifier in format string
		pad = L' ';
		sign = L'\0';
		size = L'\0';
		left = 0;
		hash = 0;
		blank = 0;
	//	group = 0;
		width = 0;
		length = 0;
		precision = -1;

		do			// flag(s) in format string?
			if (*format == L'-')
				left = 1;
			else if (*format == L'+')
				sign = L'+';
			else if (*format == L'0')
				pad = L'0';
			else if (*format == L' ')
				blank = 1;
			else if (*format == L'#')
				hash = 1;
		//	else if (*format == L'\'')
		//		group = 1;
			else
				break;
		while (*++format != L'\0');

		if (blank && sign == L'\0')
			sign = L' ';

		if (*format >= L'0'	// field width in format string?
		 && *format <= L'9')
		{
			for (width = *format++ - L'0';
			     *format >= L'0' && *format <= L'9';
			     width *= 10, width += *format++ - L'0')
				continue;

			if (width < 0)
				width = 2147483647;
		}
		else if (*format == L'*')
		{
			format++;

			width = va_arg(input, int);

			if (width < 0)
			{
				if (width == ~2147483647)
					width = 0;
				else
					width = -width;

				left = 1;
			}
		}

	//	if (left)
	//		pad = L' ';

		if (*format == L'.')	// precision in format string?
		{
			pad = L' ';

			if (*++format == L'*')
			{
				format++;

				precision = va_arg(input, int);

			//	if (precision < 0)
			//		precision = -1;
			}
			else if (*format >= L'0' && *format <= L'9')
			{
				for (precision = *format++ - L'0';
				     *format >= L'0' && *format <= L'9';
				     precision *= 10, precision += *format++ - L'0')
					continue;

				if (precision < 0)
					precision = 2147483647;
			}
			else
				precision = 0;
		}

		if (*format == L'j'	// size modifier in format string?
		 || *format == L't'
		 || *format == L'z')
			size = *format++;
		else if (*format == L'h'
		      || *format == L'l')
		{
			size = *format++;

			if (size == *format)
			{
				format++;
					// encode L"hh" as L'H' and L"ll" as L'L'
				size ^= L'a' ^ L'A';
			}
		}

		if (*format == L'd'	// conversion type in format string?
		 || *format == L'i')
		{			// signed decimal integer:
					//  [ ' ' ... | '+' | '-' ] [ '0' ... ] [ <digit> ... ] [ ' ' ... ]
			if (size == L'j'
			 || size == L't'
			 || size == L'z'
			 || size == L'L')
				(long long) ull = va_arg(input, long long);
			else if (size == L'h')
				(long long) ull = va_arg(input, short);
			else if (size == L'H')
				(long long) ull = va_arg(input, char);
			else // L'l', L't', L'z', L'\0'
				(long long) ull = va_arg(input, int);

			if ((long long) ull < 0)
			{
				ull = 0 - ull;
				sign = L'-';
			}
					// convert to decimal characters,
					//  least significant digit in scratch[0],
					//   most significant digit in scratch[length - 1]
			do
				scratch[length++] = L'0' + (wchar_t) (ull % 10);
			while (ull /= 10);
					// index of decimal integer (leading L'0' padding)
			index = length;

			if (precision > length)
				precision -= length;
			else
				precision = 0;

			if (sign != L'\0')
				scratch[length++] = sign;

			goto NUMBER;
		}
		else if (*format == L'u')
		{			// unsigned decimal integer:
					//  [{ ' ' | '0' } ... ] [ <digit> ... ] [ ' ' ... ]
			if (size == L'j'
			 || size == L't'
			 || size == L'z'
			 || size == L'L')
				ull = va_arg(input, unsigned long long);
			else if (size == L'h')
				ull = va_arg(input, unsigned short);
			else if (size == L'H')
				ull = va_arg(input, unsigned char);
			else // L'l', L't', L'z', L'\0'
				ull = va_arg(input, unsigned int);

					// convert to decimal characters,
					//  least significant digit in scratch[0],
					//   most significant digit in scratch[length - 1]
			do
				scratch[length++] = L'0' + (wchar_t) (ull % 10);
			while (ull /= 10);
					// index of decimal integer (leading L'0' padding)
			index = length;

			if (precision > length)
				precision -= length;
			else
				precision = 0;

			goto NUMBER;
		}
		else if (*format == L'o')
		{			// unsigned octal integer:
					//  [{ ' ' | '0' } ... ] [ <digit> ... ] [ ' ' ... ]
			if (size == L'j'
			 || size == L't'
			 || size == L'z'
			 || size == L'L')
				ull = va_arg(input, unsigned long long);
			else if (size == L'h')
				ull = va_arg(input, unsigned short);
			else if (size == L'H')
				ull = va_arg(input, unsigned char);
			else // L'l', L't', L'z', L'\0'
				ull = va_arg(input, unsigned int);

			if (ull == 0)
				hash = 0;
			do
				scratch[length++] = L'0' + (wchar_t) (ull % 8);
			while (ull /= 8);
					// index of octal integer (leading L'0' padding)
			index = length;

			if (hash)
#if 0
				if (precision >= 0)
					precision++;
#else
				precision += (precision >= 0);
#endif
			if (precision > length)
				precision -= length;
			else
				precision = 0;
					// prefix L'0'
			if (hash)
				scratch[length++] = L'0';

			goto NUMBER;
		}
#if 0
		else if (*format == L'x'
		      || *format == L'X')
#else
		else if ((*format | L'x' ^ L'X') == L'x')
#endif
		{			// unsigned hexadecimal integer:
					//  [ ' ' ... ] [ '0' { 'x' | 'X' }] [ '0' ... ] [ <digit> ... ] [ ' ' ... ]

			unicode = L"0123456789ABCDEF0123456789abcdef" + 16 * (*format ^ L'X') / (L'x' ^ L'X');

			if (size == L'j'
			 || size == L't'
			 || size == L'z'
			 || size == L'L')
				ull = va_arg(input, unsigned long long);
			else if (size == L'h')
				ull = va_arg(input, unsigned short);
			else if (size == L'H')
				ull = va_arg(input, unsigned char);
			else // L'l', L't', L'z', L'\0'
				ull = va_arg(input, unsigned int);

			if (ull == 0)
				hash = 0;
			do
				scratch[length++] = unicode[ull % 16];
			while (ull /= 16);
					// index of hexadecimal integer (leading L'0' padding)
			index = length;

			if (precision > length)
				precision -= length;
			else
				precision = 0;

			if (hash)
			{		// prefix L"0x" or L"0X"
				scratch[length++] = *format;
				scratch[length++] = L'0';
			}

		NUMBER:	if (width > precision + length)
				width -= precision + length;
			else
				width = 0;

			if (left)
				while (length > index && output < end)
					*output++ = scratch[--length];
			else if (pad == L'0')
			{
				while (length > index && output < end)
					*output++ = scratch[--length];
#ifndef OPTIMIZE
				while (width-- > 0 && output < end)
					*output++ = L'0';
#else
				if (width > end - output)
					width = end - output;

				// https://msdn.microsoft.com/en-us/library/1fdeehz6.aspx

				wmemset(output, L'0', width);

				output += width;
#endif // OPTIMIZE
			}
			else
			{
#ifndef OPTIMIZE
				while (width-- > 0 && output < end)
					*output++ = L' ';
#else
				if (width > end - output)
					width = end - output;

				// https://msdn.microsoft.com/en-us/library/1fdeehz6.aspx

				wmemset(output, L' ', width);

				output += width;
#endif // OPTIMIZE
				while (length > index && output < end)
					*output++ = scratch[--length];
			}
#ifndef OPTIMIZE
			while (precision-- > 0 && output < end)
				*output++ = L'0';
#else
			if (precision > end - output)
				precision = end - output;

			// https://msdn.microsoft.com/en-us/library/1fdeehz6.aspx

			wmemset(output, L'0', precision);

			output += precision;
#endif // OPTIMIZE
			while (length > 0 && output < end)
				*output++ = scratch[--length];
#ifndef OPTIMIZE
		//	if (left)
		RIGHT:		while (width-- > 0 && output < end)
					*output++ = L' ';
#else
		RIGHT:	if (left)
			{
				if (width > end - output)
					width = end - output;

				// https://msdn.microsoft.com/en-us/library/1fdeehz6.aspx

				wmemset(output, L' ', width);

				output += width;
			}
#endif // OPTIMIZE
		}
#if 0
		else if (*format == L'p'
		      || *format == L'P')
#else
		else if ((*format | L'p' ^ L'P') == L'p')
#endif
		{			// unsigned hexadecimal pointer:
					//  [ ' ' ... ] [ '0' { 'x' | 'X' }] [ '0' ... ] [ <digit> ... ] [ ' ' ... ]
			if (size != L'\0'
			 || sign != L'\0'
			 || pad != L' '
			 || precision != -1)
				goto INVALID;

			unicode = L"0123456789ABCDEF0123456789abcdef" + 16 * (*format ^ L'P') / (L'p' ^ L'P');

			ull = va_arg(input, unsigned long long);

			if (ull == 0)
				hash = 0;
			do
				scratch[length++] = unicode[ull % 16];
			while (ull /= 16);

			while (length < 2 * sizeof(ull))
				scratch[length++] = L'0';

			if (hash)
			{		// prefix L"0x" or L"0X"
				scratch[length++] = *format ^ L'P' | L'X';
				scratch[length++] = L'0';
			}

			if (!left)
				while (width-- > length && output < end)
					*output++ = L' ';

			while (length > 0 && output < end)
				*output++ = scratch[--length];

			goto RIGHT;
		}
		else if (*format == L'c')
		{			// ANSI or UNICODE character:
					//  [ ' ' ... ] [ <character> ] [ ' ' ... ]
			if (precision != -1)
				goto INVALID;

			if (size == L'h')
				goto C;	// ANSI character?

			if (size != L'\0'
			 && size != L'l')
				goto INVALID;

		c:	cw = va_arg(input, wchar_t);

		COMMON:	if (cw != L'\0')
#if 0
				if (width > 0)
					width--;
#else
				width -= (width > 0);
#endif
#ifndef OPTIMIZE
			if (!left)
				while (width-- > 0 && output < end)
					*output++ = L' ';
#else
			if (!left)
			{
				if (width > end - output)
					width = end - output;

				// https://msdn.microsoft.com/en-us/library/1fdeehz6.aspx

				wmemset(output, L' ', width);

				output += width;
			}
#endif // OPTIMIZE
			if (cw != L'\0' && output < end)
				*output++ = cw;

			goto RIGHT;
		}
		else if (*format == L'C')
		{			// ANSI or UNICODE character:
					//  [ ' ' ... ] [ <character> ] [ ' ' ... ]
			if (precision != -1)
				goto INVALID;

			if (size == L'l')
				goto c;	// UNICODE character?

			if (size != L'\0'
			 && size != L'h')
				goto INVALID;

		C:	cb = va_arg(input, char);
			cw = UCS2[cb];

			goto COMMON;
		}
		else if (*format == L's')
		{			// ANSI or UNICODE character string:
					//  [ ' ' ... ] [ <character> ... ] [ ' ' ... ]
			if (size == L'h')
				goto S;	// ANSI character string?

			if (size != L'\0'
			 && size != L'l')
				goto INVALID;

		s:	unicode = va_arg(input, wchar_t *);

			if (unicode == NULL)
				unicode = L"(null)";
#if 0
			// https://msdn.microsoft.com/en-us/library/78zh94ax.aspx

			length = wcslen(unicode);

			if (precision >= 0 && length > precision)
				length = precision;
#else
			if (precision < 0)

				// https://msdn.microsoft.com/en-us/library/78zh94ax.aspx

				length = wcslen(unicode);
			else
				for (length = 0; length < precision; length++)
					if (unicode[length] == L'\0')
						break;
#endif
		STRING:	if (width > length)
				width -= length;
			else
				width = 0;
#ifndef OPTIMIZE
			if (!left)
				while (width-- > 0 && output < end)
					*output++ = L' ';

			while (length-- > 0 && output < end)
				*output++ = *unicode++;
#else
			if (!left)
			{
				if (width > end - output)
					width = end - output;

				// https://msdn.microsoft.com/en-us/library/1fdeehz6.aspx

				wmemset(output, L' ', width);

				output += width;
			}

			if (length > end - output)
				length = end - output;

			// https://msdn.microsoft.com/en-us/library/dswaw1wk.aspx

			wmemcpy(output, unicode, length);

			output += length;
#endif // OPTIMIZE
			goto RIGHT;
		}
		else if (*format == L'S')
		{			// ANSI or UNICODE character string:
					//  [ ' ' ... ] [ <character> ... ] [ ' ' ... ]
			if (size == L'l')
				goto s;	// UNICODE character string?

			if (size != L'\0'
			 && size != L'h')
				goto INVALID;

		S:	ansi = va_arg(input, char *);

			if (ansi == NULL)
				ansi = "(null)";
#if 0
			// https://msdn.microsoft.com/en-us/library/78zh94ax.aspx

			length = strlen(ansi);

			if (precision >= 0 && length > precision)
				length = precision;
#else
			if (precision < 0)

				// https://msdn.microsoft.com/en-us/library/78zh94ax.aspx

				length = strlen(ansi);
			else
				for (length = 0; length < precision; length++)
					if (ansi[length] == '\0')
						break;
#endif
			if (width > length)
				width -= length;
			else
				width = 0;
#ifndef OPTIMIZE
			if (!left)
				while (width-- > 0 && output < end)
					*output++ = L' ';
#else
			if (!left)
			{
				if (width > end - output)
					width = end - output;

				// https://msdn.microsoft.com/en-us/library/1fdeehz6.aspx

				wmemset(output, L' ', width);

				output += width;
			}
#endif // OPTIMIZE
			while (length-- > 0 && output < end)
				*output++ = UCS2[*ansi++];

			goto RIGHT;
		}
		else if (*format == L'n')
		{			// assign character count
			if (size == L'j'
			 || size == L't'
			 || size == L'z'
			 || size == L'L')
				*va_arg(input, long long *) = (long long) (output - begin);
			else if (size == L'h')
				*va_arg(input, short *) = (short) (output - begin);
			else if (size == L'H')
				*va_arg(input, char *) = (char) (output - begin);
			else
				*va_arg(input, int *) = (int) (output - begin);
		}
#if 0
		else if (*format == L'a'
		      || *format == L'A')
#else
		else if ((*format | L'a' ^ L'A') == L'a')
#endif
		{			// hexadecimal double-precision floating-point:
					//  [ ' ' ... | '+' | '-' ] '0' { 'x' | 'X' } [ '0' ... ] <digit> [ '.' <digit> ... ] { 'p' | 'P' } { '+' | '-' } <digit> ... [ ' ' ... ]
			if (size != L'\0')
				goto INVALID;

			unicode = L"0123456789ABCDEF0123456789abcdef" + 16 * (*format ^ L'A') / (L'a' ^ L'A');
			ull = va_arg(input, unsigned long long);

			if ((long long) ull < 0)
			{		// strip sign bit
				ull &= ~0ULL >> 1;
				sign = L'-';
			}
					// biased exponent
			exponent = (int) (ull >> 52);

			if (exponent == 2047)
			{
				if (sign != L'\0')
					scratch[length++] = sign;
#ifndef OPTIMIZE
				if (ull & ~0ULL >> 12)
				{
					scratch[length++] = *format ^ L'A' | L'N';
					scratch[length++] = *format ^ L'A' | L'A';
					scratch[length++] = *format ^ L'A' | L'N';
				}
				else
				{
					scratch[length++] = *format ^ L'A' | L'I';
					scratch[length++] = *format ^ L'A' | L'N';
					scratch[length++] = *format ^ L'A' | L'F';
				}

				scratch[length] = L'\0';
#else
				// https://msdn.microsoft.com/en-us/library/dswaw1wk.aspx

				if (ull & ~0ULL >> 12)
					wmemcpy(scratch + length, *format == L'a' ? L"nan" : L"NAN", sizeof("NAN"));
				else
					wmemcpy(scratch + length, *format == L'a' ? L"inf" : L"INF", sizeof("INF"));

				length += 3;
#endif // OPTIMIZE
				unicode = scratch;

				goto STRING;
			}
			else if (exponent > 0)
			{		// unbiased exponent
				exponent -= 1023;
					// strip exponent, insert hidden bit
				ull &= ~0ULL >> 12;
				ull |= 1ULL << 52;
			}
			else		// 0.0 or denormal floating-point number
#ifndef OPTIMIZE
				if (ull)
				{	// unbiased exponent
					exponent = 1 - 1023;
					// normalize significand
					while (ull < 1ULL << 52)
						ull += ull, exponent--;
				}
#else
				// https://msdn.microsoft.com/en-us/library/fbxyd7zd.aspx

				if (_BitScanReverse64(&ul, ull))
				{	// unbiased exponent
					exponent = 1 - 1023 - (52 - ul);
					// normalize significand
					ull <<= 52 - ul;
				}
#endif // OPTIMIZE
					// convert significand to hexadecimal, one digit per byte,
					//  least significant digit in decimal[length - 1],
					//   most significant digit in decimal[0],
					//    hexadecimal point after most significant digit
			do
				decimal[length++] = (char) (ull >> 52);
			while (ull &= ~0ULL >> 12, ull <<= 4);

			point = 0;
			extra = 0;

			if (precision < 0)
				precision = length - 1;
			else if (precision > length - 1)
			{		// number of additional trailing zeroes
				extra = precision - length + 1;
				precision = length - 1;
				hash = 1;
			}
			else if (precision < length - 1
			      && (decimal[precision + 1] > 8U
			       || decimal[precision + 1] == 8U && decimal[precision] % 2U
			       || tiebreak(decimal, precision + 2, length)))
			{		// round to nearest, ties to even
#if 0
				index = precision + 1;
				digit = 8U;

				do
					digit += decimal[index],
					decimal[index] = digit % 16U;
				while (digit /= 16U && index-- > 0);

				if (digit)
				{	// hexadecimal overflow: increment exponent
					exponent++;
					// set leading digit
					decimal[0] = digit;
				}
#else
				index = precision;

				while (index >= 0 && decimal[index] == 15U)
					decimal[index--] = 0U;

				if (index < 0)
				{	// hexadecimal overflow: increment exponent
					exponent++;
					// set leading digit
					decimal[0] = 1U;
				}
				else
					decimal[index] += 1U;
#endif
			}
					// hexadecimal exponent
			if (exponent < 0)
			{
				exponent = -exponent;
				size = L'-';
			}
			else if (exponent > 0)
				size = L'+';
		//	else
		//		size = L'\0';

			length = 0;

			do		// hexadecimal exponent
				scratch[length++] = L'0' + (wchar_t) (exponent % 10U);
			while (exponent /= 10U);
					// exponent sign
			if (size != L'\0')
				scratch[length++] = size;
					// exponent prefix L'p' or L'P'
			scratch[length++] = *format ^ L'A' | L'P';
					// index of exponent prefix (trailing L'0' padding)
			exponent = length;
					// hexadecimal point
			if (precision > 0 || hash > 0)
				scratch[length++] = L'.';
					// most significant digit
			scratch[length++] = L'0' + decimal[0];
					// index of most significant digit (leading L'0' padding)
			index = length;
					// prefix L"0x" or L"0X"
			scratch[length++] = *format ^ L'A' | L'X';
			scratch[length++] = L'0';
					// sign
			if (sign != L'\0')
				scratch[length++] = sign;

			goto DOUBLE;
		}
#if 0
		else if (*format == L'e'
		      || *format == L'E')
#else
		else if ((*format | L'e' ^ L'E') == L'e')
#endif
		{			// decimal double-precision floating-point:
					//  [ ' ' ... | '+' | '-' ] [ '0' ... ] <digit> [ '.' <digit> ... [ '0' ... ]] { 'e' | 'E' } [ '+' | '-' ] <digit> ... [ ' ' ... ]
			if (size != L'\0')
				goto INVALID;

			unicode = NULL;

			ull = va_arg(input, unsigned long long);

			if ((long long) ull < 0)
			{		// strip sign bit
				ull &= ~0ULL >> 1;
				sign = L'-';
			}
					// biased exponent
			exponent = (int) (ull >> 52);

			if (exponent == 2047)
			{
				if (sign != L'\0')
					scratch[length++] = sign;
#ifndef OPTIMIZE
				if (ull & ~0ULL >> 12)
				{
					scratch[length++] = *format ^ L'E' | L'N';
					scratch[length++] = *format ^ L'E' | L'A';
					scratch[length++] = *format ^ L'E' | L'N';
				}
				else
				{
					scratch[length++] = *format ^ L'E' | L'I';
					scratch[length++] = *format ^ L'E' | L'N';
					scratch[length++] = *format ^ L'E' | L'F';
				}

				scratch[length] = L'\0';
#else
				// https://msdn.microsoft.com/en-us/library/dswaw1wk.aspx

				if (ull & ~0ULL >> 12)
					wmemcpy(scratch + length, *format == L'e' ? L"nan" : L"NAN", sizeof("NAN"));
				else
					wmemcpy(scratch + length, *format == L'e' ? L"inf" : L"INF", sizeof("INF"));

				length += 3;
#endif // OPTIMIZE
				unicode = scratch;

				goto STRING;
			}
			else if (exponent > 0)
			{		// unbiased exponent
				exponent -= 1023;
					// strip exponent, insert hidden bit
				ull &= ~0ULL >> 12;
				ull |= 1ULL << 52;
			}
			else		// 0.0 or denormal floating-point number
#ifndef OPTIMIZE
				if (ull)
				{	// unbiased exponent
					exponent = 1 - 1023;
					// normalize significand
					while (ull < 1ULL << 52)
						ull += ull, exponent--;
				}
#else
				// https://msdn.microsoft.com/en-us/library/fbxyd7zd.aspx

				if (_BitScanReverse64(&ul, ull))
				{	// unbiased exponent
					exponent = 1 - 1023 - (52 - ul);
					// normalize significand
					ull <<= 52 - ul;
				}
#endif // OPTIMIZE
					// convert significand to decimal, one digit per byte,
					//  least significant digit in decimal[length - 1],
					//   most significant digit in decimal[0],
					//    decimal point after most significant digit:
					//     0x0                 = 0[.]
					//     0x1[.]0000000000000 = 1[.]
					//     0x1[.]0000000000001 = 1[.]0000000000000002220446049250313080847263336181640625
					//     0x1[.]8000000000000 = 1[.]5
					//     0x1[.]FFFFFFFFFFFFF = 1[.]9999999999999997779553950749686919152736663818359375
			do
				decimal[length++] = (char) (ull >> 52);
			while (ull &= ~0ULL >> 12, ull *= 10);

					// decimal point past unhidden bit
					//  (also decimal exponent)
			point = 0;
					// multiply by 2**exponent
			while (exponent > 0)
			{
				index = length;
				digit = 0U;

				if (decimal[0] < 5U)
				{
					do
						digit += decimal[index - 1] * 2U,
						decimal[index - 1] = digit % 10U, digit /= 10U;
					while (--index > 0);
				}
				else	// decimal overflow: move decimal point and all digits
				{	//  one place right, insert new non-zero leading digit
					point++;
					length++;

					do
						digit += decimal[index - 1] * 2U,
						decimal[index] = digit % 10U, digit /= 10U;
					while (--index > 0);

					decimal[0] = digit;
				}

				exponent--;
			}
					// divide by 2**-exponent
			while (exponent < 0)
			{
				index = 0;

				if (decimal[length - 1] % 2U)
					decimal[length++] = 0U;

				if (decimal[0] > 1U)
				{
					digit = 0U;

					do
						digit = digit * 10U + decimal[index],
						decimal[index] = digit / 2U, digit %= 2U;
					while (++index < length);
				}
				else	// decimal underflow: discard new leading 0, move
				{	//  decimal point and all digits one place left
					point--;
					length--;
					digit = decimal[0];

					do
						digit = digit * 10U + decimal[index + 1],
						decimal[index] = digit / 2U, digit %= 2U;
					while (++index < length);
				}

				exponent++;
			}

			if (precision < 0)
				precision = 6;

			extra = 0;

			if (precision > length - 1)
			{		// number of additional trailing zeroes
				extra = precision - length + 1;
				precision = length - 1;
				hash = 1;
			}
			else if (precision < length - 1
			      && (decimal[precision + 1] > 5U
			       || decimal[precision + 1] == 5U && decimal[precision] % 2U
			       || tiebreak(decimal, precision + 2, length)))
			{		// round to nearest, ties to even
#if 0
				index = precision + 1;
				digit = 5U;

				do
					digit += decimal[index],
					decimal[index] = digit % 10U;
				while (digit /= 10U && index-- > 0);

				if (digit)
				{	// decimal overflow: move decimal point right
					point++;
					// set leading digit
					decimal[0] = digit;
				}
#else
				index = precision;

				while (index >= 0 && decimal[index] == 9U)
					decimal[index--] = 0U;

				if (index < 0)
				{	// decimal overflow: move decimal point right
					point++;
					// set leading digit
					decimal[0] = 1U;
				}
				else
					decimal[index] += 1U;
#endif
			}
					// decimal exponent
			if (point < 0)
			{
				point = -point;
				size = L'-';
			}
			else if (point > 0)
				size = L'+';
		//	else
		//		size = L'\0';

			do		// decimal exponent
				scratch[exponent++] = L'0' + (wchar_t) (point % 10U);
			while (point /= 10U);
					// exponent sign
			if (size != L'\0')
				scratch[exponent++] = size;
					// exponent prefix L'e' or L'E'
			scratch[exponent++] = *format;
					// index of exponent prefix (trailing L'0' padding)
			length = exponent;
					// decimal point
			if (precision > 0 || hash > 0)
				scratch[length++] = L'.';
					// most significant digit
			scratch[length++] = L'0' + decimal[0];
					// index of most significant digit (leading L'0' padding)
			index = length;
					// sign
			if (sign != L'\0')
				scratch[length++] = sign;

		DOUBLE:	if (width > length + precision + extra)
				width -= length + precision + extra;
			else
				width = 0;

			if (left)
			{		// sign, prefix, most significant digit and point
				while (length > exponent && output < end)
					*output++ = scratch[--length];

				if (unicode == NULL)
					while (point < precision && output < end)
						*output++ = L'0' + decimal[++point];
				else
					while (point < precision && output < end)
						*output++ = unicode[decimal[++point]];
#ifndef OPTIMIZE
				while (extra-- > 0 && output < end)
					*output++ = L'0';
#else
				if (extra > end - output)
					extra = end - output;

				// https://msdn.microsoft.com/en-us/library/1fdeehz6.aspx

				wmemset(output, L'0', extra);

				output += extra;
#endif // OPTIMIZE
				while (length > 0 && output < end)
					*output++ = scratch[--length];

				goto RIGHT;
			}

			if (pad == L'0')
			{		// sign and prefix
				while (length > index && output < end)
					*output++ = scratch[--length];
#ifndef OPTIMIZE
				while (width-- > 0 && output < end)
					*output++ = L'0';
#else
				if (width > end - output)
					width = end - output;

				// https://msdn.microsoft.com/en-us/library/1fdeehz6.aspx

				wmemset(output, L'0', width);

				output += width;
#endif // OPTIMIZE
			}
			else
			{
#ifndef OPTIMIZE
				while (width-- > 0 && output < end)
					*output++ = L' ';
#else
				if (width > end - output)
					width = end - output;

				// https://msdn.microsoft.com/en-us/library/1fdeehz6.aspx

				wmemset(output, L' ', width);

				output += width;
#endif // OPTIMIZE
				while (length > index && output < end)
					*output++ = scratch[--length];
			}

			while (length > exponent && output < end)
				*output++ = scratch[--length];

			if (unicode == NULL)
				while (point < precision && output < end)
					*output++ = L'0' + decimal[++point];
			else
				while (point < precision && output < end)
					*output++ = unicode[decimal[++point]];
#ifndef OPTIMIZE
			while (extra-- > 0 && output < end)
				*output++ = L'0';
#else
			if (extra > end - output)
				extra = end - output;

			// https://msdn.microsoft.com/en-us/library/1fdeehz6.aspx

			wmemset(output, L'0', extra);

			output += extra;
#endif // OPTIMIZE
			while (length > 0 && output < end)
				*output++ = scratch[--length];
		}
		else			// return negative index of invalid character
		INVALID:		//  in format string
			return (int) (start - format);

		format++;
	}

	if (output != NULL)
		*output = L'\0';

	return (int) (output - begin);	// return length of output string
}

// https://msdn.microsoft.com/en-us/library/ybk95axf.aspx

int	swprintf(wchar_t *output, size_t count, const wchar_t *format, ...)
{
	int	result;

	// https://msdn.microsoft.com/en-us/library/kb57fad8.aspx

	va_list	input;
	va_start(input, format);

	// https://msdn.microsoft.com/en-us/library/28d5ce15.aspx

	result = vswprintf(output, count, format, input);

	va_end(input);

	return result;
}
<<NOKEEP

AMD64-DLG.OBJ: AMD64-MSC.MAK
	ML64.EXE /Brepro /c /Fo$@ /nologo /Ta<< /W3 /X
; Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

; https://msdn.microsoft.com/en-us/library/7cycf5bc.aspx

	extern	__pfnDliDefaultHook2 :ptr proc
	extern	__pfnDliFailureHook2 (__pfnDliDefaultHook2) :ptr proc
	extern	__pfnDliNotifyHook2 (__pfnDliDefaultHook2) :ptr proc
	end
<<NOKEEP

AMD64-DLH.OBJ: AMD64-MSC.MAK
	CL.EXE /Brepro /c /Fo$@ /GAFwy /nologo /O2txy /Tc<< /W4 /wd4201 /wd4204 /Zl
// Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

// https://msdn.microsoft.com/en-us/library/7f0aews7.aspx

#pragma comment(compiler)
#pragma comment(lib, "kernel32.lib")
#if 0
#pragma comment(linker, "/ALTERNATENAME:__pfnDliFailureHook2=__pfnDliDefaultHook2")
#pragma comment(linker, "/ALTERNATENAME:__pfnDliNotifyHook2=__pfnDliDefaultHook2")
#endif
#pragma comment(user, "(C)opyright 2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>")

#define STRICT
#define WIN32_LEAN_AND_MEAN

#include <windows.h>

#ifndef EXCEPTION_DELAY_LOAD_INVALID_PARAMETER
#define EXCEPTION_DELAY_LOAD_INVALID_PARAMETER	0xC06D0057
#endif

// https://msdn.microsoft.com/en-us/library/1c9e046h.aspx

#ifndef EXCEPTION_DELAY_LOAD_MODULE_NOT_FOUND
#define EXCEPTION_DELAY_LOAD_MODULE_NOT_FOUND	0xC06D007E
#endif

#ifndef EXCEPTION_DELAY_LOAD_ENTRY_NOT_FOUND
#define EXCEPTION_DELAY_LOAD_ENTRY_NOT_FOUND	0xC06D007F
#endif

extern	const	IMAGE_DOS_HEADER	__ImageBase;

typedef	DWORD	RVA;

// https://msdn.microsoft.com/en-us/library/f0fkfy9y.aspx
// https://msdn.microsoft.com/en-us/library/16b2dyk5.aspx

typedef	enum	dliNotify
{
	dliStartProcessing,
	dliNotePreLoadLibrary,
	dliNotePreGetProcAddress,
	dliFailLoadLib,
	dliFailGetProc,
	dliNoteEndProcessing
} dliNotify;

typedef	struct	DelayLoadDescr
{
	DWORD	dwAttributes;		// 1UL = all members are RVAs
	DWORD	dwDllName;
	DWORD	dwHMODULE;		// RVA of module handle
	DWORD	dwIAT;			// RVA of import address table
	DWORD	dwINT;			// RVA of import name table
	DWORD	dwBoundIAT;		// RVA of optional bound import address table
	DWORD	dwUnloadIAT;		// RVA of optional copy of original import address table
	DWORD	dwTimeStamp;
} DelayLoadDescr;

typedef	struct	DelayLoadProc
{
	BOOL	fImportByName;
	union
	{
		LPCSTR	szProcName;
		DWORD	dwOrdinal;
	};
} DelayLoadProc;

typedef	struct	DelayLoadInfo
{
	DWORD		cb;		// size of structure
	DelayLoadDescr	*pidd;		// raw form of data (everything is there)
	FARPROC		*ppfn;		// points to address of function to load
	LPCSTR		szDll;		// name of DLL
	DelayLoadProc	dlp;		// name or ordinal of function to load
	HMODULE		hmodCur;	// handle of DLL
	FARPROC		pfnCur;		// actual function that will be called
	DWORD		dwLastError;	// error received (if an error notification)
} DelayLoadInfo;

typedef	FARPROC	(WINAPI *PfnDliHook) (dliNotify, DelayLoadInfo *);

extern	PfnDliHook	__pfnDliNotifyHook2;
extern	PfnDliHook	__pfnDliFailureHook2;
const	PfnDliHook	__pfnDliDefaultHook2 = NULL;

// https://msdn.microsoft.com/en-us/library/2b054ds4.aspx

BOOL	WINAPI	__FUnloadDelayLoadedDLL2(LPCSTR szDll);

// https://msdn.microsoft.com/en-us/library/09t6x5ds.aspx

FARPROC	WINAPI	__delayLoadHelper2(DelayLoadDescr *lpDLD, FARPROC *lpfnIATEntry)
{
	HMODULE	hModule;

	// https://msdn.microsoft.com/en-us/library/ff1shykw.aspx

	HMODULE			*lpHMODULE = (HMODULE *) ((LPBYTE) &__ImageBase + lpDLD->dwHMODULE);
	IMAGE_THUNK_DATA	*lpINT = (IMAGE_THUNK_DATA *) ((LPBYTE) &__ImageBase + lpDLD->dwINT);
	IMAGE_THUNK_DATA	*lpIAT = (IMAGE_THUNK_DATA *) ((LPBYTE) &__ImageBase + lpDLD->dwIAT);
	IMAGE_THUNK_DATA	*lpBoundIAT = (IMAGE_THUNK_DATA *) ((LPBYTE) &__ImageBase + lpDLD->dwBoundIAT);
	DWORD			dwEntry = (IMAGE_THUNK_DATA *) lpfnIATEntry - lpIAT;

	// NOTE: *lpfnIATEntry == lpIAT[dwEntry].u1.Function

	DelayLoadInfo	dli = {sizeof(DelayLoadInfo),
			       lpDLD,
			       lpfnIATEntry,
			       (LPCSTR) &__ImageBase + lpDLD->dwDllName,
			       {!IMAGE_SNAP_BY_ORDINAL(lpINT[dwEntry].u1.Ordinal),
			         IMAGE_SNAP_BY_ORDINAL(lpINT[dwEntry].u1.Ordinal)
			       ? IMAGE_ORDINAL(lpINT[dwEntry].u1.Ordinal)
			       : ((IMAGE_IMPORT_BY_NAME *) ((LPBYTE) &__ImageBase + lpINT[dwEntry].u1.AddressOfData))->Name},
			       *lpHMODULE,
			       (FARPROC) NULL,
			       ERROR_SUCCESS};

	if (__pfnDliNotifyHook2 != NULL)
	{
		dli.pfnCur = (*__pfnDliNotifyHook2)(dliStartProcessing, &dli);

		if (dli.pfnCur != NULL)
			goto SUCCESS;
	}

	if (lpDLD->dwAttributes != 0UL)
	{
		dli.dwLastError = ERROR_INVALID_PARAMETER;

		// https://msdn.microsoft.com/en-us/library/ms680552.aspx
		// https://msdn.microsoft.com/en-us/library/1c9e046h.aspx

		RaiseException(EXCEPTION_DELAY_LOAD_INVALID_PARAMETER,
		               EXCEPTION_NONCONTINUABLE,
		               1UL,
		               (DWORD_PTR *) &dli);

		goto FAILURE;
	}

	if (dli.hmodCur != NULL)	// module already loaded?
		goto ADDRESS;

	if (__pfnDliNotifyHook2 != NULL)
		dli.hmodCur = (HMODULE) (*__pfnDliNotifyHook2)(dliNotePreLoadLibrary, &dli);

	if (dli.hmodCur != NULL)	// module handle resolved by notification routine?
		goto ADDRESS;
#ifndef LOAD_LIBRARY_SEARCH_SYSTEM32
	// https://msdn.microsoft.com/en-us/library/ms684175.aspx

	dli.hmodCur = LoadLibraryA(dli.szDll);
#else
	// https://msdn.microsoft.com/en-us/library/ms684179.aspx

	dli.hmodCur = LoadLibraryExA(dli.szDll, NULL, LOAD_LIBRARY_SEARCH_SYSTEM32);
#endif
	if (dli.hmodCur == NULL)	// module not loaded?
	{
		// https://msdn.microsoft.com/en-us/library/ms679360.aspx

		dli.dwLastError = GetLastError();

		if (__pfnDliFailureHook2 != NULL)
			dli.hmodCur = (HMODULE) (*__pfnDliFailureHook2)(dliFailLoadLib, &dli);

		if (dli.hmodCur == NULL)
		{
			// https://msdn.microsoft.com/en-us/library/ms680552.aspx
			// https://msdn.microsoft.com/en-us/library/1c9e046h.aspx

			RaiseException(EXCEPTION_DELAY_LOAD_MODULE_NOT_FOUND,
			               EXCEPTION_NONCONTINUABLE,
			               1UL,
			               (DWORD_PTR *) &dli);

			goto FAILURE;
		}

		// https://msdn.microsoft.com/en-us/library/853x471w.aspx

		hModule = (HMODULE) _InterlockedExchangePointer((LPVOID *) lpHMODULE, dli.hmodCur);

		if (hModule == dli.hmodCur)

			// https://msdn.microsoft.com/en-us/library/ms683152.aspx

			FreeLibrary(dli.hmodCur);
		else
			if (lpDLD->dwUnloadIAT != 0UL)
			{
				//...
			}
	}
ADDRESS:
	if (__pfnDliNotifyHook2 != NULL)
		dli.pfnCur = (*__pfnDliNotifyHook2)(dliNotePreGetProcAddress, &dli);

	if (dli.pfnCur != NULL)		// function address resolved by notification routine?
		goto SUCCESS;

	if ((lpBoundIAT != NULL) && (lpDLD->dwTimeStamp != 0UL))
	{
		// https://msdn.microsoft.com/en-us/library/ms680336.aspx

		IMAGE_NT_HEADERS	*lpModule = (IMAGE_NT_HEADERS *) ((LPBYTE) dli.hmodCur + ((IMAGE_DOS_HEADER *) dli.hmodCur)->e_lfanew);

		if ((lpModule->Signature == IMAGE_NT_SIGNATURE)
		 && (lpModule->FileHeader.TimeDateStamp == lpDLD->dwTimeStamp)
		 && (lpModule->OptionalHeader.ImageBase == dli.hmodCur))
		{
			dli.pfnCur = (FARPROC) lpBoundIAT[dwEntry].u1.Function;

			if (dli.pfnCur != NULL)
				goto SUCCESS;
		}
	}

	// https://msdn.microsoft.com/en-us/library/ms683212.aspx

	dli.pfnCur = GetProcAddress(dli.hmodCur, dli.dlp.szProcName);

	if (dli.pfnCur != NULL)		// function address found?
		goto SUCCESS;

	// https://msdn.microsoft.com/en-us/library/ms679360.aspx

	dli.dwLastError = GetLastError();

	if (__pfnDliFailureHook2 != NULL)
		dli.pfnCur = (*__pfnDliFailureHook2)(dliFailGetProc, &dli);

	if (dli.pfnCur != NULL)		// function address resolved by failure routine?
		goto SUCCESS;

	// https://msdn.microsoft.com/en-us/library/ms680552.aspx
	// https://msdn.microsoft.com/en-us/library/1c9e046h.aspx

	RaiseException(EXCEPTION_DELAY_LOAD_ENTRY_NOT_FOUND,
	               EXCEPTION_NONCONTINUABLE,
	               1UL,
	               (DWORD_PTR *) &dli);

	goto FAILURE;

SUCCESS:
	*lpfnIATEntry = dli.pfnCur;
FAILURE:
	if (__pfnDliNotifyHook2 != NULL)
		(*__pfnDliNotifyHook2)(dliNoteEndProcessing, &dli);

	return dli.pfnCur;
}
<<NOKEEP

AMD64-DLI.OBJ: AMD64-MSC.MAK
	CL.EXE /Brepro /c /Fo$@ /GAFwy /nologo /O2txy /Tc<< /W4 /wd4201 /wd4204 /Zl
// Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

// https://msdn.microsoft.com/en-us/library/7f0aews7.aspx

#pragma comment(compiler)
#pragma comment(lib, "kernel32.lib")
#pragma comment(user, "(C)opyright 2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>")

#define STRICT
#define WIN32_LEAN_AND_MEAN

#include <windows.h>

#ifndef EXCEPTION_DELAY_LOAD_INVALID_PARAMETER
#define EXCEPTION_DELAY_LOAD_INVALID_PARAMETER	0xC06D0057
#endif

// https://msdn.microsoft.com/en-us/library/1c9e046h.aspx

#ifndef EXCEPTION_DELAY_LOAD_MODULE_NOT_FOUND
#define EXCEPTION_DELAY_LOAD_MODULE_NOT_FOUND	0xC06D007E
#endif

#ifndef EXCEPTION_DELAY_LOAD_ENTRY_NOT_FOUND
#define EXCEPTION_DELAY_LOAD_ENTRY_NOT_FOUND	0xC06D007F
#endif

extern	const	IMAGE_DOS_HEADER	__ImageBase;

typedef	DWORD	RVA;

// https://msdn.microsoft.com/en-us/library/f0fkfy9y.aspx
// https://msdn.microsoft.com/en-us/library/16b2dyk5.aspx

typedef	enum	dliNotify
{
	dliStartProcessing,
	dliNotePreLoadLibrary,
	dliNotePreGetProcAddress,
	dliFailLoadLib,
	dliFailGetProc,
	dliNoteEndProcessing
} dliNotify;

typedef	struct	DelayLoadDescr
{
	DWORD	dwAttributes;		// 1UL = all members are RVAs
	DWORD	dwDllName;
	DWORD	dwHMODULE;		// RVA of module handle
	DWORD	dwIAT;			// RVA of import address table
	DWORD	dwINT;			// RVA of import name table
	DWORD	dwBoundIAT;		// RVA of optional bound import address table
	DWORD	dwUnloadIAT;		// RVA of optional copy of original import address table
	DWORD	dwTimeStamp;
} DelayLoadDescr;

typedef	struct	DelayLoadProc
{
	BOOL	fImportByName;
	union
	{
		LPCSTR	szProcName;
		DWORD	dwOrdinal;
	};
} DelayLoadProc;

typedef	struct	DelayLoadInfo
{
	DWORD		cb;		// size of structure
	DelayLoadDescr	*pidd;		// raw form of data (everything is there)
	FARPROC		*ppfn;		// points to address of function to load
	LPCSTR		szDll;		// name of DLL
	DelayLoadProc	dlp;		// name or ordinal of function to load
	HMODULE		hmodCur;	// handle of DLL
	FARPROC		pfnCur;		// actual function that will be called
	DWORD		dwLastError;	// error received (if an error notification)
} DelayLoadInfo;

typedef	FARPROC	(WINAPI *PfnDliHook) (dliNotify, DelayLoadInfo *);

// https://msdn.microsoft.com/en-us/library/2b054ds4.aspx

BOOL	WINAPI	__FUnloadDelayLoadedDLL2(LPCSTR szDll);

// https://msdn.microsoft.com/en-us/library/09t6x5ds.aspx

FARPROC	WINAPI	__delayLoadHelper2(DelayLoadDescr *lpDLD, FARPROC *lpfnIATEntry)
{
	HMODULE	hModule;

	// https://msdn.microsoft.com/en-us/library/ff1shykw.aspx

	HMODULE			*lpHMODULE = (HMODULE *) ((LPBYTE) &__ImageBase + lpDLD->dwHMODULE);
	IMAGE_THUNK_DATA	*lpINT = (IMAGE_THUNK_DATA *) ((LPBYTE) &__ImageBase + lpDLD->dwINT);
	IMAGE_THUNK_DATA	*lpIAT = (IMAGE_THUNK_DATA *) ((LPBYTE) &__ImageBase + lpDLD->dwIAT);
	IMAGE_THUNK_DATA	*lpBoundIAT = (IMAGE_THUNK_DATA *) ((LPBYTE) &__ImageBase + lpDLD->dwBoundIAT);
	DWORD			dwEntry = (IMAGE_THUNK_DATA *) lpfnIATEntry - lpIAT;

	// NOTE: *lpfnIATEntry == lpIAT[dwEntry].u1.Function

	DelayLoadInfo	dli = {sizeof(DelayLoadInfo),
			       lpDLD,
			       lpfnIATEntry,
			       (LPCSTR) &__ImageBase + lpDLD->dwDllName,
			       {!IMAGE_SNAP_BY_ORDINAL(lpINT[dwEntry].u1.Ordinal),
			         IMAGE_SNAP_BY_ORDINAL(lpINT[dwEntry].u1.Ordinal)
			       ? IMAGE_ORDINAL(lpINT[dwEntry].u1.Ordinal)
			       : ((IMAGE_IMPORT_BY_NAME *) ((LPBYTE) &__ImageBase + lpINT[dwEntry].u1.AddressOfData))->Name},
			       *lpHMODULE,
			       (FARPROC) NULL,
			       ERROR_SUCCESS};

	if (lpDLD->dwAttributes != 0UL)
	{
		dli.dwLastError = ERROR_INVALID_PARAMETER;

		// https://msdn.microsoft.com/en-us/library/ms680552.aspx
		// https://msdn.microsoft.com/en-us/library/1c9e046h.aspx

		RaiseException(EXCEPTION_DELAY_LOAD_INVALID_PARAMETER,
		               EXCEPTION_NONCONTINUABLE,
		               1UL,
		               (DWORD_PTR *) &dli);

		return (FARPROC) NULL;
	}

	if (dli.hmodCur == NULL)	// module not yet loaded?
	{
#ifndef LOAD_LIBRARY_SEARCH_SYSTEM32
		// https://msdn.microsoft.com/en-us/library/ms684175.aspx

		dli.hmodCur = LoadLibraryA(dli.szDll);
#else
		// https://msdn.microsoft.com/en-us/library/ms684179.aspx

		dli.hmodCur = LoadLibraryExA(dli.szDll, NULL, LOAD_LIBRARY_SEARCH_SYSTEM32);
#endif
		if (dli.hmodCur == NULL)
		{
			// https://msdn.microsoft.com/en-us/library/ms679360.aspx

			dli.dwLastError = GetLastError();

			// https://msdn.microsoft.com/en-us/library/ms680552.aspx
			// https://msdn.microsoft.com/en-us/library/1c9e046h.aspx

			RaiseException(EXCEPTION_DELAY_LOAD_MODULE_NOT_FOUND,
			               EXCEPTION_NONCONTINUABLE,
			               1UL,
			               (DWORD_PTR *) &dli);

			return (FARPROC) NULL;
		}

		// https://msdn.microsoft.com/en-us/library/853x471w.aspx

		hModule = (HMODULE) _InterlockedExchangePointer((LPVOID *) lpHMODULE, dli.hmodCur);

		if (hModule == dli.hmodCur)

			// https://msdn.microsoft.com/en-us/library/ms683152.aspx

			FreeLibrary(dli.hmodCur);
		else
			if (lpDLD->dwUnloadIAT != 0UL)
			{
				//...
			}
	}

	if ((lpBoundIAT != NULL) && (lpDLD->dwTimeStamp != 0UL))
	{
		// https://msdn.microsoft.com/en-us/library/ms680336.aspx

		IMAGE_NT_HEADERS	*lpModule = (IMAGE_NT_HEADERS *) ((LPBYTE) dli.hmodCur + ((IMAGE_DOS_HEADER *) dli.hmodCur)->e_lfanew);

		if ((lpModule->Signature == IMAGE_NT_SIGNATURE)
		 && (lpModule->FileHeader.TimeDateStamp == lpDLD->dwTimeStamp)
		 && (lpModule->OptionalHeader.ImageBase == dli.hmodCur))
		{
			dli.pfnCur = (FARPROC) lpBoundIAT[dwEntry].u1.Function;

			if (dli.pfnCur != NULL)
				return *lpfnIATEntry = dli.pfnCur;
		}
	}

	// https://msdn.microsoft.com/en-us/library/ms683212.aspx

	dli.pfnCur = GetProcAddress(dli.hmodCur, dli.dlp.szProcName);

	if (dli.pfnCur != NULL)		// function address resolved?
		return *lpfnIATEntry = dli.pfnCur;

	// https://msdn.microsoft.com/en-us/library/ms679360.aspx

	dli.dwLastError = GetLastError();

	// https://msdn.microsoft.com/en-us/library/ms680552.aspx
	// https://msdn.microsoft.com/en-us/library/1c9e046h.aspx

	RaiseException(EXCEPTION_DELAY_LOAD_ENTRY_NOT_FOUND,
	               EXCEPTION_NONCONTINUABLE,
	               1UL,
	               (DWORD_PTR *) &dli);

	return (FARPROC) NULL;
}
<<NOKEEP

AMD64-GS.OBJ: AMD64-MSC.MAK
	ML64.EXE /Brepro /c /Fo$@ /nologo /Ta<< /W3 /X
; Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

; https://msdn.microsoft.com/en-us/library/ms680328.aspx

_lcu_64	struct	8
	dword	sizeof _lcu_64
	dword	'VOID'		; 2006-04-21 21:32:06 UTC
	word	0, 815
	dword	0, 0, 0
	qword	6 dup (0)
	dword	0
	word	0, 2048		; LOAD_LIBRARY_SEARCH_SYSTEM32
	qword	0
	qword	offset __security_cookie
	qword	6 dup (0)
	dword	0
_lcu_64	ends

	.const

	public	_load_config_used
_load_config_used \
	_lcu_64	<>		; IMAGE_LOAD_CONFIG_DIRECTORY64

	.data

	public	__security_cookie
__security_cookie \
	qword	3141592653589793241 shr 16

	.code

__security_check_cookie \
	proc	public		; void __security_check_cookie(qword cookie)

	cmp	rcx, __security_cookie
	jne	short fastfail

;;	shr	rcx, 48
;;	jnz	short fastfail

	ret

; https://msdn.microsoft.com/en-us/library/dn774154.aspx

fastfail:
	mov	ecx, 2		; rcx = FAST_FAIL_STACK_COOKIE_CHECK_FAILURE
	int	41
	ud2

__security_check_cookie \
	endp

__security_init_cookie \
	proc	public		; void __security_init_cookie(void)

	mov	rax, __security_cookie
	mov	rcx, 3141592653589793241 shr 16
	cmp	rcx, rax
	je	short init

	test	rax, rax
	jne	short exit
init:
	rdtsc			; rax = low dword of time stamp counter,
				; rdx = high dword of time stamp counter
	mov	ecx, edx	; rcx = high dword of time stamp counter
	bswap	edx
	imul	rcx, rax	; rcx = high dword of time stamp counter
				;     * low dword of time stamp counter
	bswap	rax
	xor	rax, rdx	; rax = byte-swapped time stamp counter
	mul	rcx
	xor	rax, rdx	; rax = random number
	shr	rax, 16
	mov	__security_cookie, rax
exit:
	ret

__security_init_cookie \
	endp

__GSHandlerCheck \
	proc	private		; int __GSHandlerCheck(void *, void *, void *, void *)

	xor	eax, eax
	inc	eax		; rax = ExceptionContinueSearch
	ret

__GSHandlerCheck \
	endp
	end
<<NOKEEP

AMD64-MEM.OBJ: AMD64-MSC.MAK
	ML64.EXE /Brepro /c /Fo$@ /nologo /Ta<< /W3 /X
; Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

; Microsoft calling convention for AMD64 platform:
; - first 4 integer or pointer arguments (from left to right) are passed
;   in registers RCX/R1, RDX/R2, R8 and R9;
; - 16-byte arguments are passed by reference;
; - surplus arguments are pushed on stack in reverse order (from right
;   to left), 8-byte aligned;
; - caller allocates memory for 16-byte return value and passes pointer
;   to it as (hidden) first argument, thus shifting all other arguments;
; - caller always allocates "home space" for 4 arguments on stack, even
;   when less than 4 arguments are passed, but does not need to push
;   first 4 arguments;
; - callee can spill first 4 arguments from registers to "home space";
; - callee can clobber "home space";
; - stack is 16-byte aligned: callee must decrement RSP by 8+n*16 bytes
;   when it calls other functions (CALL instruction pushes 8 bytes);
; - 64-bit integer or pointer result is returned in register RAX/R0;
; - 32-bit integer result is returned in register EAX;
; - registers RAX/R0, RCX/R1, RDX/R2, R8, R9, R10 and R11 are volatile
;   and can be clobbered;
; - registers RBX/R3, RSP/R4, RBP/R5, RSI/R6, RDI/R7, R12, R13, R14 and
;   R15 must be preserved.

	.code

; https://msdn.microsoft.com/en-us/library/d7zdhf37.aspx

memchr	proc	public		; void *memchr(void const *destination, int character, size_t count)

	mov	r9, rdi
	mov	rdi, rcx	; rdi = address of destination
	mov	rcx, r8		; rcx = count
	mov	eax, edx	; rax = character
	repne	scasb
	lea	rax, [rdi-1]
	test	rcx, rcx
	cmovz	rax, rcx	; rax = (rcx = 0) ? 0 : address of character
	mov	rdi, r9
	ret

memchr	endp

; https://msdn.microsoft.com/en-us/library/zyaebf12.aspx

memcmp	proc	public		; int memcmp(void const *source, void const *destination, size_t count)

;;	xor	eax, eax	; rax = 0
;;	test	r8, r8
;;	jz	short equal	; count = 0?

;;	cmp	rcx, rdx
;;	je	short equal	; address of source = address of destination?

	mov	r9, rsi
	mov	rsi, rcx	; rsi = address of source
	mov	rcx, r8		; rcx = count
	mov	r8, rdi
	mov	rdi, rdx	; rdi = address of destination
	xor	eax, eax	; rax = 0,
				; CF = 0,
				; ZF = 1 (required when count is 0)
	repe	cmpsb
	seta	al		; rax = (*source > *destination)
	sbb	rax, 0		; rax = (*source > *destination)
				;     - (*source < *destination)
				;     = {1, 0, -1}
	mov	rdi, r8
	mov	rsi, r9
equal:
	ret

memcmp	endp

; https://msdn.microsoft.com/en-us/library/dswaw1wk.aspx

memcpy	proc	public		; void *memcpy(void *destination, void const *source, size_t count)

	mov	rax, rcx	; rax = address of destination
;;	test	r8, r8
;;	jz	short return	; count = 0?

;;	cmp	rcx, rdx
;;	je	short return	; address of destination = address of source?

	mov	r9, rdi
	mov	rdi, rcx	; rdi = address of destination
	mov	rcx, r8		; rcx = count
	mov	r8, rsi
	mov	rsi, rdx	; rsi = address of source
	rep	movsb
	mov	rdi, r9
	mov	rsi, r8
return:
	ret

memcpy	endp

memmem	proc	public		; void *memmem(void const *haystack, size_t count,
				;              void const *needle, size_t length)

	xor	eax, eax	; rax = address of needle in haystack = 0
	test	rdx, rdx
	jz	short empty	; length of haystack = 0?

	cmp	rdx, r9
	jb	short empty	; length of haystack < length of needle?

	mov	rax, rcx	; rax = address of haystack
	test	r9, r9
	jz	short empty	; length of needle = 0?

	mov	r10, rdi
	mov	rdi, rcx	; rdi = address of haystack
	mov	rcx, rdx	; rcx = length of haystack
	mov	r11, rsi
search:
	mov	al, [r8]	; al = first character of needle
				; rdi = current address in haystack
	repne	scasb		; rdi = next address in haystack,
				; rcx = current length of haystack
	jne	short break	; (first character of) needle not in haystack?

	dec	rcx		; rcx = next length of haystack

	mov	al, [r8+r9-1]	; al = last character of needle
	cmp	al, [rdi+r9-2]
	jne	short continue	; last character of needle not in haystack?
compare:
	mov	rax, rdi	; rax = next address in haystack
	mov	rdx, rcx	; rdx = next length of haystack
if 0
	dec	rdi		; rdi = current address in haystack
				;     = address of matching character
	mov	rsi, r8		; rsi = address of needle
	mov	rcx, r9		; rcx = length of needle
else
				; rdi = next address in haystack
	mov	rsi, r8
	inc	rsi		; rsi = address of needle + 1
	mov	rcx, r9
	dec	rcx		; rcx = length of needle - 1,
				; ZF = (rcx = 0)
;;	jz	short match	; needle in haystack?
endif
	repe	cmpsb
	je	short match	; needle in haystack?

	mov	rdi, rax	; rdi = current address in haystack
	mov	rcx, rdx	; rcx = current length of haystack
continue:
	cmp	rcx, r9
	jae	short search	; length of haystack >= length of needle?
break:
	xor	eax, eax
	mov	rdi, r10
	mov	rsi, r11
empty:
	ret
match:
	dec	rax		; rax = address of needle in haystack
	mov	rdi, r10
	mov	rsi, r11
	ret

memmem	endp

; https://msdn.microsoft.com/en-us/library/8k35d1fx.aspx

memmove	proc	public		; void *memmove(void *destination, void const *source, size_t count)

	mov	rax, rcx	; rax = address of destination
;;	test	r8, r8
;;	jz	short return	; count = 0?

	cmp	rcx, rdx
	je	short return	; address of destination = address of source?

	mov	r9, rdi
	mov	rdi, rcx	; rdi = address of destination
	mov	rcx, r8		; rcx = count
	mov	r8, rsi
	mov	rsi, rdx	; rsi = address of source

	jb	short default	; address of destination < address of source?

	add	rdx, rcx	; rdx = address of source + count
	cmp	rdi, rdx
	jae	short default	; address of destination >= address of source + count?
overlap:
	lea	rdi, [rdi+rcx-1]
	lea	rsi, [rsi+rcx-1]
	std
default:
	rep	movsb
	cld
	mov	rdi, r9
	mov	rsi, r8
return:
	ret

memmove	endp

memrchr	proc	public		; void *memrchr(void const *destination, int character, size_t count)

	mov	r9, rdi
	lea	rdi, [rcx+r8-1]	; rdi = address of destination + count - 1
	mov	eax, edx	; rax = character
	mov	rcx, r8		; rcx = count
	std
	repne	scasb
	cld
	lea	rax, [rdi+1]
	test	rcx, rcx
	cmovz	rax, rcx	; rax = (rcx = 0) ? 0 : address of character
	mov	rdi, r9
	ret

memrchr	endp

; https://msdn.microsoft.com/en-us/library/1fdeehz6.aspx

memset	proc	public		; void *memset(void *destination, int character, size_t count)

	mov	r9, rcx		; r9 = address of destination
	mov	rcx, r8		; rcx = count
;;	jrcxz	short @f	; count = 0?

	mov	r8, rdi
	mov	rdi, r9		; rdi = address of destination
	mov	eax, edx	; rax = character
	rep	stosb
	mov	rdi, r8
@@:
	mov	rax, r9		; rax = address of destination
	ret

memset	endp
	end
<<NOKEEP

AMD64-RTC.OBJ: AMD64-MSC.MAK

AMD64-STK.OBJ: AMD64-MSC.MAK
	ML64.EXE /Brepro /c /Fo$@ /nologo /Ta<< /W3 /X
; Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

; https://msdn.microsoft.com/en-us/library/dn449700.aspx
; https://msdn.microsoft.com/en-us/library/ms253960.aspx

_nt_tib	struct	8		; thread information block
chain	qword	?		; address of first exception registration record
base	qword	?		; stack base
limit	qword	?		; stack limit
	qword	?		; address of subsystem thread information block
fiber	qword	?		; fiber data
pointer	qword	?		; arbitrary user pointer
self	qword	?		; address of _nt_tib
_nt_tib	ends

	.code

; MSC internal intrinsic _alloca() alias __chkstk():
; receives argument in rax

; https://docs.microsoft.com/en-us/windows/win32/devnotes/-win32-__chkstk

; NOTE: _alloca() must preserve rax and all argument registers;
;       it can raise 'stack overflow' exception!

;;	alias	<_alloca> = <__chkstk>

__chkstk proc	public		; qword __chkstk(qword size)

	mov	r10, gs:[_nt_tib.limit]
				; r10 = (current) stack limit
	lea	r11, [rsp+8]	; r11 = stack pointer of caller
	sub	r11, rax	; r11 = new stack pointer
	jnb	short limit
overflow:
	xor	r11, r11	; r11 = 0
probe:
	sub	r10, 4096	; r10 = address of guard page
	test	r10, [r10]	; r10 = new stack limit via 'guard page' exception
limit:
	cmp	r10, r11
	ja	short probe	; stack limit > new stack pointer?

	ret

__chkstk endp
	end
<<NOKEEP

AMD64-STR.OBJ: AMD64-MSC.MAK
	ML64.EXE /Brepro /c /Fo$@ /nologo /Ta<< /W3 /X
; Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

; NOTE: characters are unsigned!

	.code

; https://msdn.microsoft.com/en-us/library/h1x0y282.aspx

strcat	proc	public		; char *strcat(char *destination, char const *source)

	mov	r9, rcx		; r9 = address of destination string
	mov	r10, rdi
ifdef VARIANT
	mov	rdi, rcx	; rdi = address of destination string
	xor	eax, eax	; rax = '\0'
ifdef AMD
	stc
	sbb	rcx, rcx	; rcx = -1
else
	or	rcx, -1		; rcx = -1
endif
	repne	scasb
	dec	rdi		; rdi = address of '\0'
				;     = end of destination string
	mov	r11, rsi
	mov	rsi, rdi	; rsi = end of destination string
	mov	rdi, rdx	; rdi = address of source string
;;	xor	eax, eax
ifdef AMD
	stc
	sbb	rcx, rcx	; rcx = -1
else
	or	rcx, -1		; rcx = -1
endif
	repne	scasb
	not	rcx		; rcx = length of source string (including '\0')
	mov	rdi, rsi	; rdi = end of destination string
	mov	rsi, rdx	; rsi = address of source string
else ; VARIANT
	mov	rdi, rdx	; rdi = address of source string
;;	xor	eax, eax
ifdef AMD
	stc
	sbb	rcx, rcx	; rcx = -1
else
	or	rcx, -1		; rcx = -1
endif
	repne	scasb
	not	rcx		; rcx = length of source string (including '\0')
	mov	r11, rsi
	mov	rsi, rdx	; rsi = address of source string
	mov	rdx, rcx
	mov	rdi, r9		; rdi = address of destination string
;;	xor	eax, eax	; rax = '\0'
ifdef AMD
	stc
	sbb	rcx, rcx	; rcx = -1
else
	or	rcx, -1		; rcx = -1
endif
	repne	scasb
	dec	rdi		; rdi = address of '\0'
				;     = end of destination string
	mov	rcx, rdx	; rcx = length of source string (including '\0')
endif ; VARIANT
	rep	movsb
	mov	rax, r9		; rax = address of destination string
	mov	rdi, r10
	mov	rsi, r11
	ret

strcat	endp

; https://msdn.microsoft.com/en-us/library/b34ccac3.aspx

strchr	proc	public		; char *strchr(char const *string, int character)

	mov	r10, rdi
	mov	rdi, rcx	; rdi = address of string
	xor	eax, eax	; rax = '\0'
ifdef AMD
	stc
	sbb	rcx, rcx	; rcx = -1
else
	or	rcx, -1		; rcx = -1
endif
	repne	scasb
	not	rcx		; rcx = length of string (including '\0')
	mov	rax, rdx	; rax = character
	sub	rdi, rcx	; rdi = address of string
	repne	scasb
	lea	rax, [rdi-1]	; rax = address of character
	cmovne	rax, rcx	; rax = (rcx = 0) ? 0 : address of character
	mov	rdi, r10
	ret

strchr	endp

; https://msdn.microsoft.com/en-us/library/e0z9k731.aspx

strcmp	proc	public		; ssize_t strcmp(char const *source, char const *destination)

	xor	eax, eax	; rax = 0
	cmp	rcx, rdx
	je	short equal	; address of source string = address of destination string?

	mov	r11, rsi
	mov	rsi, rcx	; rsi = address of source string
	mov	r10, rdi
	mov	rdi, rdx	; rdi = address of destination string
;;	xor	eax, eax	; rax = '\0'
ifdef AMD
	stc
	sbb	rcx, rcx	; rcx = -1
else
	or	rcx, -1		; rcx = -1
endif
	repne	scasb
	not	rcx		; rcx = length of destination string (including '\0')
	mov	rdi, rdx	; rdi = address of destination string
;;	xor	eax, eax	; rax = 0
	repe	cmpsb
	seta	al		; rax = (*source > *destination)
	sbb	rax, 0		; rax = (*source > *destination)
				;     - (*source < *destination)
				;     = {1, 0, -1}
	mov	rdi, r10
	mov	rsi, r11
equal:
	ret

strcmp	endp

; https://msdn.microsoft.com/en-us/library/a7cwbx4t.aspx

; NOTE: strcoll() is another implementation of strcmp()!

strcoll	proc	public		; ssize_t strcoll(char const *source, char const *destination)

	sub	rdx, rcx
	jz	short equal	; address of destination string = address of source string?
compare:
	mov	al, [rcx]
	cmp	al, [rcx+rdx]
	jne	short different

	inc	rcx
	test	al, al
	jnz	short compare	; *source <> '\0'?
equal:
	xor	eax, eax	; rax = 0
	ret

different:
	sbb	rax, rax	; rax = (*source < *destination) ? -1 : 0
	or	rax, 1		; rax = (*source < *destination)
				;     - (*source > *destination)
				;     = {-1, 0, 1}
	ret

strcoll	endp

; https://msdn.microsoft.com/en-us/library/kk6xf663.aspx

strcpy	proc	public		; char *strcpy(char *destination, char const *source)

	mov	r9, rcx		; r9 = address of destination string
	mov	r10, rdi
	mov	rdi, rdx	; rdi = address of source string
	xor	eax, eax	; rax = '\0'
ifdef AMD
	stc
	sbb	rcx, rcx	; rcx = -1
else
	or	rcx, -1		; rcx = -1
endif
	repne	scasb
	not	rcx		; rcx = length of source string (including '\0')
	mov	rdi, r9		; rdi = address of destination string
	mov	r11, rsi
	mov	rsi, rdx	; rsi = address of source string
	rep	movsb
	mov	rax, r9		; rax = address of destination string
	mov	rdi, r10
	mov	rsi, r11
	ret

strcpy	endp

; https://msdn.microsoft.com/en-us/library/xe8sk0x7.aspx

strcspn	proc	public		; size_t strcspn(char const *string, char const *delimiter)

	xor	eax, eax	; rax = 0
	cmp	al, [rdx]
	je	short empty	; delimiter[0] = '\0'?

	mov	[rsp+32], rax
	mov	[rsp+24], rax
	mov	[rsp+16], rax
	mov	[rsp+8], rax	; bitmap[0..255] = 0
setup:
	bts	[rsp+8], rax	; bitmap[rax] = 1
	mov	al, [rdx]	; rax = *delimiter
	inc	rdx		; rdx = ++delimiter
	cmp	al, ah
	jne	short setup	; rax <> '\0'?

	mov	rdx, rcx	; rdx = address of string
skip:
	mov	al, [rcx]	; rax = *string
	inc	rcx		; rcx = ++string
	bt	[rsp+8], rax
	jnc	short skip	; bitmap[rax] = 0 (no match)?
stop:
	sbb	rcx, rdx	; rcx = number of non-matching characters
	mov	rax, rcx
	ret
empty:
	mov	rdx, rcx	; rdx = address of string
count:
	cmp	al, [rcx]
	lea	rcx, [rcx+1]	; rcx = ++string
	jne	short count	; *string <> '\0'?

	stc
	sbb	rcx, rdx	; rcx = number of characters
	mov	rax, rcx
	ret

strcspn	endp

; https://msdn.microsoft.com/en-us/library/78zh94ax.aspx

strlen	proc	public		; size_t strlen(char const *string)

	mov	rdx, rdi
	mov	rdi, rcx	; rdi = address of string
	xor	eax, eax	; rax = '\0'
ifdef AMD
	stc
	sbb	rcx, rcx	; rcx = -1
else
	or	rcx, -1		; rcx = -1
endif
	repne	scasb
if 0
	not	rcx		; rcx = length of string (including '\0')
	dec	rcx
	mov	rax, rcx	; rax = length of string
else
	mov	rax, -2
	sub	rax, rcx	; rax = length of string
endif
	mov	rdi, rdx
	ret

strlen	endp

; https://msdn.microsoft.com/en-us/library/tbyd7s1y.aspx

strncat	proc	private		; char *strncat(char *destination, char const *source, size_t count)

	ud2

strncat	endp

; https://msdn.microsoft.com/en-us/library/eywx8zcx.aspx

strncmp	proc	private		; int strncmp(char const *source, char const *destination, size_t count)

	ud2

strncmp	endp

; https://msdn.microsoft.com/en-us/library/xdsywd25.aspx

strncpy	proc	private		; char *strncpy(char *destination, char const *source, size_t count)

	ud2

strncpy	endp

; https://msdn.microsoft.com/en-us/library/z50ty2zh.aspx

strnlen	proc	private		; size_t strnlen(char const *string, size_t count)

	ud2

strnlen	endp

; https://msdn.microsoft.com/en-us/library/ms235495.aspx

strnset	proc	private		; char *strnset(char const *string, int character, size_t count)

	ud2

strnset	endp

; https://msdn.microsoft.com/en-us/library/217yyhy9.aspx

strpbrk	proc	public		; char *strpbrk(char const *string, char const *delimiter)

	xor	eax, eax	; rax = 0
	cmp	al, [rdx]
	je	short empty	; delimiter[0] = '\0'?

	mov	[rsp+32], rax
	mov	[rsp+24], rax
	mov	[rsp+16], rax
	mov	[rsp+8], rax	; bitmap[0..255] = 0
setup:
	bts	[rsp+8], rax	; bitmap[rax] = 1
	mov	al, [rdx]	; rax = *delimiter
	inc	rdx		; rdx = ++delimiter
	cmp	al, ah
	jne	short setup	; rax <> '\0'?
skip:
	mov	al, [rcx]	; rax = *string
	inc	rcx		; rcx = ++string
	bt	[rsp+8], rax
	jnc	short skip	; bitmap[rax] = 0 (no match)?
stop:
	dec	rcx		; rcx = --string
	neg	eax
	sbb	rax, rax	; rax = (*string = '\0') ? 0 : -1
	and	rax, rcx	; rax = (*string = '\0') ? 0 : address of string
empty:
	ret

strpbrk	endp

; https://msdn.microsoft.com/en-us/library/ftw0heb9.aspx

strrchr	proc	public		; char *strrchr(char const *string, int character)

	mov	r10, rdi
	mov	rdi, rcx	; rdi = address of string
	xor	eax, eax	; rax = '\0'
ifdef AMD
	stc
	sbb	rcx, rcx	; rcx = -1
else
	or	rcx, -1		; rcx = -1
endif
	repne	scasb
	not	rcx		; rcx = length of string (including '\0')
	mov	rax, rdx	; rax = character
	dec	rdi		; rdi = address of '\0'
				;     = end of string
	std
	repne	scasb
	cld
	lea	rax, [rdi+1]	; rax = address of character
	cmovne	rax, rcx	; rax = (rcx = 0) ? 0 : address of character
	mov	rdi, r10
	ret

strrchr	endp

; https://msdn.microsoft.com/en-us/library/ms235404.aspx

strrev	proc	private		; char *strrev(char *string)

	ud2

strrev	endp

; https://msdn.microsoft.com/en-us/library/ms235359.aspx

strset	proc	public		; char *strset(char *string, int character)

	mov	r9, rcx		; r9 = address of string
	mov	r10, rdi
	mov	rdi, rcx	; rdi = address of string
	xor	eax, eax	; rax = '\0'
ifdef AMD
	stc
	sbb	rcx, rcx	; rcx = -1
else
	or	rcx, -1		; rcx = -1
endif
	repne	scasb
	not	rcx		; rcx = length of string (including '\0')
	dec	rcx
	mov	rdi, r9		; rdi = address of string
	mov	rax, rdx	; rax = character
	rep	stosb
	mov	rax, r9		; rax = address of string
	mov	rdi, r10
	ret

strset	endp

; https://msdn.microsoft.com/en-us/library/kthzzay9.aspx

strspn	proc	public		; size_t strspn(char const *string, char const *delimiter)

	xor	eax, eax	; rax = 0
	cmp	al, [rdx]
	je	short empty	; delimiter[0] = '\0'?

	mov	[rsp+32], rax
	mov	[rsp+24], rax
	mov	[rsp+16], rax
	mov	[rsp+8], rax	; bitmap[0..255] = 0

	mov	al, [rdx]	; rax = *delimiter
	inc	rdx		; rdx = ++delimiter
setup:
	bts	[rsp+8], rax	; bitmap[rax] = 1
	mov	al, [rdx]	; rax = *delimiter
	inc	rdx		; rdx = ++delimiter
	cmp	al, ah
	jne	short setup	; rax <> '\0'?

	mov	rdx, rcx	; rdx = address of string
skip:
	mov	al, [rcx]	; rax = *string
	inc	rcx		; rcx = ++string
	bt	[rsp+8], rax
	jc	short skip	; bitmap[rax] = 1 (match)?
if 0
	dec	rcx		; rcx = --string
	sub	rcx, rdx
else
	stc
	sbb	rcx, rdx	; rcx = number of matching characters
endif
	mov	rax, rcx
empty:
	ret

strspn	endp

; https://msdn.microsoft.com/en-us/library/z9da80kz.aspx

strstr	proc	public		; char *strstr(char const *haystack, char const *needle)

	mov	r8, rcx		; r8 = address of haystack
	mov	r10, rdi
	mov	rdi, rdx	; rdi = address of needle
	xor	eax, eax	; rax = '\0'
ifdef AMD
	stc
	sbb	rcx, rcx	; rcx = -1
else
	or	rcx, -1		; rcx = -1
endif
	repne	scasb
	not	rcx		; rcx = length of needle (including '\0')
	dec	rcx		; rcx = length of needle
	mov	rax, r8		; rax = address of haystack
	jz	short empty	; length of needle = 0?

	mov	r9, rcx		; r9 = length of needel
	mov	rdi, r8		; rdi = address of haystack
	xor	eax, eax	; rax = '\0'
ifdef AMD
	stc
	sbb	rcx, rcx	; rcx = -1
else
	or	rcx, -1		; rcx = -1
endif
	repne	scasb
	not	rcx		; rcx = length of haystack (including '\0')
	sub	rdi, rcx	; rdi = address of haystack
	dec	rcx		; rcx = length of haystack
	jz	short empty	; length of haystack = 0?

	cmp	rcx, r9
	jb	short empty	; length of haystack <length of needle?

	mov	r11, rsi
search:
	mov	al, [rdx]	; al = first character of needle
				; rdi = current address in haystack
	repne	scasb		; rdi = next address in haystack,
				; rcx = current length of haystack
	jne	short break	; (first character of) needle not in haystack?

	dec	rcx		; rcx = next length of haystack

	mov	al, [rdx+r9-1]	; al = last character of needle
	cmp	al, [rdi+r9-2]
	jne	short continue	; last character of needle not in haystack?
compare:
	mov	rax, rdi	; rax = next address in haystack
	mov	r8, rcx		; r8 = next length of haystack
if 0
	dec	rdi		; rdi = current address in haystack
				;     = address of matching character
	mov	rsi, rdx	; rsi = address of needle
	mov	rcx, r9		; rcx = length of needle
else
				; rdi = next address in haystack
	mov	rsi, rdx
	inc	rsi		; rsi = address of needle + 1
	mov	rcx, r9
	dec	rcx		; rcx = length of needle - 1,
				; ZF = (rcx = 0)
;;	jz	short match	; needle in haystack?
endif
	repe	cmpsb
	je	short match	; needle in haystack?

	mov	rdi, rax	; rdi = current address in haystack
	mov	rcx, r8		; rcx = current length of haystack
continue:
	cmp	rcx, r9
	jae	short search	; length of haystack >= length of needle?
break:
	xor	eax, eax
	mov	rdi, r10
	mov	rsi, r11
empty:
	ret
match:
	dec	rax		; rax = address of needle in haystack
	mov	rdi, r10
	mov	rsi, r11
	ret

strstr	endp

strtok_r proc	public		; char *strtok_r(char *string, char const *delimiter, char **next)

	xor	eax, eax	; rax = 0
	test	rcx, rcx
	jnz	short start	; string <> 0?

	or	rcx, [r8]	; rcx = next
	jz	short null	; address of next = 0 = address of string?
start:
	cmp	al, [rcx]
	je	short null	; string[0] = '\0'?

	cmp	al, [rdx]
	je	short empty	; *delimiter = '\0'?

	mov	[rsp+32], rax
	mov	[rsp+24], rax
	mov	[rsp+16], rax
	mov	[rsp+8], rax	; bitmap[0..255] = 0

	mov	al, [rdx]	; rax = *delimiter
	inc	rdx		; rdx = ++delimiter
setup:
	bts	[rsp+8], rax	; bitmap[rax] = 1
	mov	al, [rdx]	; rax = *delimiter
	inc	rdx		; rdx = ++delimiter
	cmp	al, ah
	jne	short setup	; rax <> '\0'?
skip:
	mov	al, [rcx]	; rax = *string
	inc	rcx		; rcx = ++string
	bt	[rsp+8], rax
	jc	short skip	; bitmap[rax] = 1 (rax is a delimiter)?

	cmp	al, ah
	je	short none	; rax = '\0'?

	mov	al, ah		; rax = 0
	bts	[rsp+8], rax	; bitmap['\0'] = 1
	lea	rdx, [rcx-1]	; rdx = address of token
token:
	mov	al, [rcx]	; rax = *string
	inc	rcx		; rcx = ++string
	bt	[rsp+8], rax
	jnc	short token	; bitmap[rax] = 0 (rax is not a delimiter)?

	cmp	al, ah
	je	short last	; rax = '\0'?

	mov	[rcx-1], ah	; string[-1] = '\0' (terminate token)
	mov	[r8], rcx	; *next = address of string
	mov	rax, rdx	; rax = address of token
	ret
last:
	mov	[r8], rax	; *next = 0
	mov	rax, rdx	; rax = address of token
	ret
empty:
	mov	[r8], rax	; *next = 0
	mov	rax, rcx	; rax = address of token
	ret
null:
none:
	mov	[r8], rax	; *next = 0
	ret

strtok_r endp
	end
<<NOKEEP

AMD64-TLS.OBJ: AMD64-MSC.MAK
	ML64.EXE /Brepro /c /Fo$@ /nologo /Ta<< /W3 /X
; Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

; https://msdn.microsoft.com/en-us/library/ms680547.aspx

_tls_64	struct	8
	qword	offset _tls_begin
	qword	offset _tls_end
	qword	offset _tls_index
	qword	offset _tls_start
	dword	'VOID'		; BUG: the module loader does NOT support the 'SizeOfZeroFill' member!
	dword	0
_tls_64	ends

_bss	segment	alias(".bss$$T") dword read write 'BSS'
	public	_tls_index
_tls_index dword ?		; assigned by the module loader!
_bss	ends

_note	segment	alias(".comment") discard info read 'INFO'
	byte	"(C)opyright 2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>"
_note	ends

; https://msdn.microsoft.com/en-us/library/wxz26dz2.aspx
; https://msdn.microsoft.com/en-us/library/sf9b18xk.aspx

_linker	segment alias(".drectve") discard info read 'INFO'
	byte	"/MERGE:.ptr=.rdata /SECTION:.tls,!w"
_linker	ends

_start	segment alias(".ptr") para read 'CONST'
_tls_start equ	$$
_start	ends

_stop	segment alias(".ptr$$~~~") read 'CONST'
	qword	0		; callback function array terminator
_stop	ends

_rdata	segment alias(".rdata$$T") para read 'CONST'
	public	_tls_used
_tls_used _tls_64 <>		; IMAGE_TLS_DIRECTORY64
_rdata	ends

_begin	segment alias(".tls") para read write 'DATA'
_tls_begin equ	$$
_begin	ends

_end	segment alias(".tls$$~~~") byte read write 'DATA'
_tls_end equ	$$
_end	ends
	end
<<NOKEEP

AMD64-WCS.OBJ: AMD64-MSC.MAK
	ML64.EXE /Brepro /c /Fo$@ /nologo /Ta<< /W3 /X
; Copyright  2004-2024, Stefan Kanthak <stefan.kanthak@nexgo.de>

; NOTE: counts and lengths are numbers of wide characters, not bytes!

	.code

; https://msdn.microsoft.com/en-us/library/h1x0y282.aspx

wcscat	proc	public		; wchar_t *wcscat(wchar_t *destination, wchar_t const *source)

	mov	r9, rcx		; r9 = address of destination string
	mov	r10, rdi
ifdef VARIANT
	mov	rdi, rcx	; rdi = address of destination string
	xor	eax, eax	; rax = L'\0'
ifdef AMD
	stc
	sbb	rcx, rcx	; rcx = -1
else
	or	rcx, -1		; rcx = -1
endif
	repne	scasw
	dec	rdi		; rdi = address of L'\0'
				;     = end of destination string
	mov	r11, rsi
	mov	rsi, rdi	; rsi = end of destination string
	mov	rdi, rdx	; rdi = address of source string
;;	xor	eax, eax
ifdef AMD
	stc
	sbb	rcx, rcx	; rcx = -1
else
	or	rcx, -1		; rcx = -1
endif
	repne	scasw
	not	rcx		; rcx = length of source string (including L'\0')
	mov	rdi, rsi	; rdi = end of destination string
	mov	rsi, rdx	; rsi = address of source string
else ; VARIANT
	mov	rdi, rdx	; rdi = address of source string
;;	xor	eax, eax
ifdef AMD
	stc
	sbb	rcx, rcx	; rcx = -1
else
	or	rcx, -1		; rcx = -1
endif
	repne	scasw
	not	rcx		; rcx = length of source string (including L'\0')
	mov	r11, rsi
	mov	rsi, rdx	; rsi = address of source string
	mov	rdx, rcx
	mov	rdi, r9		; rdi = address of destination string
;;	xor	eax, eax	; rax = L'\0'
ifdef AMD
	stc
	sbb	rcx, rcx	; rcx = -1
else
	or	rcx, -1		; rcx = -1
endif
	repne	scasw
	dec	rdi
	dec	rdi		; rdi = address of L'\0'
				;     = end of destination string
	mov	rcx, rdx	; rcx = length of source string (including L'\0')
endif ; VARIANT
	rep	movsw
	mov	rax, r9		; rax = address of destination string
	mov	rdi, r10
	mov	rsi, r11
	ret

wcscat	endp

; https://msdn.microsoft.com/en-us/library/b34ccac3.aspx

wcschr	proc	public		; wchar_t *wcschr(wchar_t const *string, wchar_t character)

	mov	r10, rdi
	mov	rdi, rcx	; rdi = address of string
	xor	eax, eax	; rax = L'\0'
ifdef AMD
	stc
	sbb	rcx, rcx	; rcx = -1
else
	or	rcx, -1		; rcx = -1
endif
	repne	scasw
	not	rcx		; rcx = length of string (including L'\0')
	mov	rax, rdx	; rax = wide character
	sub	rdi, rcx	; rdi = address of string
	repne	scasw
	lea	rax, [rdi-2]	; rax = address of wide character
	cmovne	rax, rcx	; rax = (rcx = 0) ? 0 : address of wide character
	mov	rdi, r10
	ret

wcschr	endp

; https://msdn.microsoft.com/en-us/library/e0z9k731.aspx

wcscmp	proc	public		; int wcscmp(wchar_t const *source, wchar_t const *destination)

	xor	eax, eax	; rax = 0
	cmp	rcx, rdx
	je	short equal	; address of source string = address of destination string?

	mov	r11, rsi
	mov	rsi, rcx	; rsi = address of source string
	mov	r10, rdi
	mov	rdi, rdx	; rdi = address of destination string
;;	xor	eax, eax	; rax = L'\0'
ifdef AMD
	stc
	sbb	rcx, rcx	; rcx = -1
else
	or	rcx, -1		; rcx = -1
endif
	repne	scasw
	not	rcx		; rcx = length of destination string (including L'\0')
	mov	rdi, rdx	; rdi = address of destination string
;;	xor	eax, eax	; rax = 0
	repe	cmpsw
	seta	al		; rax = (*source > *destination)
	sbb	rax, 0		; rax = (*source > *destination)
				;     - (*source < *destination)
				;     = {1, 0, -1}
	mov	rdi, r10
	mov	rsi, r11
equal:
	ret

wcscmp	endp

; https://msdn.microsoft.com/en-us/library/a7cwbx4t.aspx

; NOTE: wcscoll() is another implementation of wcscmp()!

wcscoll	proc	public		; int wcscoll(wchar_t const *source, wchar_t const *destination)

	sub	rdx, rcx
	jz	short equal	; address of destination string = address of source string?
compare:
	mov	ax, [rcx]
	cmp	ax, [rcx+rdx]
	jne	short different

	lea	rcx, [rcx+2]
	test	ax, ax
	jnz	short compare	; *source <> L'\0'?
equal:
	xor	eax, eax	; rax = 0
	ret

different:
	sbb	rax, rax	; rax = (*source < *destination) ? -1 : 0
	or	rax, 1		; rax = (*source < *destination)
				;     - (*source > *destination)
				;     = {-1, 0, 1}
	ret

wcscoll	endp

; https://msdn.microsoft.com/en-us/library/kk6xf663.aspx

wcscpy	proc	public		; wchar_t *wcscpy(wchar_t *destination, wchar_t const *source)

	mov	r9, rcx		; r9 = address of destination string
	mov	r10, rdi
	mov	rdi, rdx	; rdi = address of source string
	xor	eax, eax	; rax = L'\0'
ifdef AMD
	stc
	sbb	rcx, rcx	; rcx = -1
else
	or	rcx, -1		; rcx = -1
endif
	repne	scasw
	not	rcx		; rcx = length of source string (including L'\0')
	mov	rdi, r9		; rdi = address of destination string
	mov	r11, rsi
	mov	rsi, rdx	; rsi = address of source string
	rep	movsw
	mov	rax, r9		; rax = address of destination string
	mov	rdi, r10
	mov	rsi, r11
	ret

wcscpy	endp

; https://msdn.microsoft.com/en-us/library/78zh94ax.aspx

wcslen	proc	public		; size_t wcslen(wchar_t const *string)

	mov	rdx, rdi
	mov	rdi, rcx	; rdi = address of string
	xor	eax, eax	; rax = L'\0'
ifdef AMD
	stc
	sbb	rcx, rcx	; rcx = -1
else
	or	rcx, -1		; rcx = -1
endif
	repne	scasw
	not	rcx		; rcx = length of string (including L'\0')
	dec	rcx
	dec	rcx
	mov	rax, rcx	; rax = length of string
	mov	rdi, rdx
	ret

wcslen	endp

; https://msdn.microsoft.com/en-us/library/tbyd7s1y.aspx

wcsncat	proc	private		; wchar_t *wcsncat(wchar_t *destination, wchar_t const *source, size_t count)

	ud2

wcsncat	endp

; https://msdn.microsoft.com/en-us/library/eywx8zcx.aspx

wcsncmp	proc	private		; int wcsncmp(wchar_t const *source, wchar_t const *destination, size_t count)

	ud2

wcsncmp	endp

; https://msdn.microsoft.com/en-us/library/xdsywd25.aspx

wcsncpy	proc	private		; wchar_t *wcsncpy(wchar_t *destination, wchar_t const *source, size_t count)

	ud2

wcsncpy	endp

; https://msdn.microsoft.com/en-us/library/z50ty2zh.aspx

wcsnlen	proc	private		; size_t wcsnlen(wchar_t const *string, size_t count)

	ud2

wcsnlen	endp

; https://msdn.microsoft.com/en-us/library/ms235495.aspx

wcsnset	proc	private		; wchar_t *wcsnset(wchar_t *string, wchar_t character, size_t count)

	ud2

wcsnset	endp

; https://msdn.microsoft.com/en-us/library/ftw0heb9.aspx

wcsrchr	proc	public		; wchar_t *wcsrchr(wchar_t const *string, wchar_t character)

	mov	r10, rdi
	mov	rdi, rcx	; rdi = address of string
	xor	eax, eax	; rax = L'\0'
ifdef AMD
	stc
	sbb	rcx, rcx	; rcx = -1
else
	or	rcx, -1		; rcx = -1
endif
	repne	scasw
	not	rcx		; rcx = length of string (including L'\0')
	mov	rax, rdx	; rax = wide character
	lea	rdi, [rdi-2]	; rdi = address of L'\0'
				;     = end of string
	std
	repne	scasw
	cld
	lea	rax, [rdi+2]	; rax = address of wide character
	cmovne	rax, rcx	; rax = (rcx = 0) ? 0 : address of wide character
	mov	rdi, r10
	ret

wcsrchr	endp

; https://msdn.microsoft.com/en-us/library/ms235404.aspx

wcsrev	proc	private		; wchar_t *wcsrev(wchar_t *string)

	ud2

wcsrev	endp

; https://msdn.microsoft.com/en-us/library/ms235359.aspx

wcsset	proc	public		; wchar_t *wcsset(wchar_t *string, wchar_t character)

	mov	r9, rcx		; r9 = address of string
	mov	r10, rdi
	mov	rdi, rcx	; rdi = address of string
	xor	eax, eax	; rax = L'\0'
ifdef AMD
	stc
	sbb	rcx, rcx	; rcx = -1
else
	or	rcx, -1		; rcx = -1
endif
	repne	scasw
	not	rcx		; rcx = length of string (including L'\0')
	dec	rcx
	mov	rdi, r9		; rdi = address of string
	mov	rax, rdx	; rax = wide character
	rep	stosw
	mov	rax, r9		; rax = address of string
	mov	rdi, r10
	ret

wcsset	endp

; https://msdn.microsoft.com/en-us/library/z9da80kz.aspx

wcsstr	proc	public		; wchar_t *wcsstr(wchar_t const *haystack, wchar_t const *needle)

	mov	r8, rcx		; r8 = address of haystack
	mov	r10, rdi
	mov	rdi, rdx	; rdi = address of needle
	xor	eax, eax	; rax = L'\0'
ifdef AMD
	stc
	sbb	rcx, rcx	; rcx = -1
else
	or	rcx, -1		; rcx = -1
endif
	repne	scasw
	not	rcx		; rcx = length of needle (including L'\0')
	dec	rcx		; rcx = length of needle
	mov	rax, r8		; rax = address of haystack
	jz	short empty	; length of needle = 0?

	mov	r9, rcx		; r9 = length of needel
	mov	rdi, r8		; rdi = address of haystack
	xor	eax, eax	; rax = L'\0'
ifdef AMD
	stc
	sbb	rcx, rcx	; rcx = -1
else
	or	rcx, -1		; rcx = -1
endif
	repne	scasw
	not	rcx		; rcx = length of haystack (including L'\0')
	mov	rdi, r8		; rdi = address of haystack
	dec	rcx		; rcx = length of haystack
	jz	short empty	; length of haystack = 0?

	cmp	rcx, r9
	jb	short empty	; length of haystack <length of needle?

	mov	r11, rsi
search:
	mov	ax, [rdx]	; ax = first wide character of needle
				; rdi = current address in haystack
	repne	scasw		; rdi = next address in haystack,
				; rcx = current length of haystack
	jne	short break	; (first wide character of) needle not in haystack?

	dec	rcx		; rcx = next length of haystack

	mov	ax, [rdx+r9*2-2]
				; ax = last wide character of needle
	cmp	ax, [rdi+r9*2-4]
	jne	short continue	; last wide character of needle not in haystack?
compare:
	mov	rax, rdi	; rax = next address in haystack
	mov	r8, rcx		; r8 = next length of haystack
if 0
	lea	rdi, [rdi-2]	; rdi = current address in haystack
				;     = address of matching character
	mov	rsi, rdx	; rsi = address of needle
	mov	rcx, r9		; rcx = length of needle
else
				; rdi = next address in haystack
	lea	rsi, [rdx+2]	; rsi = address of needle + 2
	mov	rcx, r9
	dec	rcx		; rcx = length of needle - 1,
				; ZF = (rcx = 0)
;;	jz	short match	; needle in haystack?
endif
	repe	cmpsw
	je	short match	; needle in haystack?

	mov	rdi, rax	; rdi = current address in haystack
	mov	rcx, r8		; rcx = current length of haystack
continue:
	cmp	rcx, r9
	jae	short search	; length of haystack >= length of needle?
break:
	xor	eax, eax
	mov	rdi, r10
	mov	rsi, r11
empty:
	ret
match:
	lea	rax, [rax-2]	; rax = address of needle in haystack
	mov	rdi, r10
	mov	rsi, r11
	ret

wcsstr	endp
	end
<<NOKEEP
